/******************************************************************************
 * xpf_dags.c:	insertion, deletion, search routines for maintaining the set  *
 *		of associations between patterns and infos (modeled on	      *
 *		PAT trees). They are VERY complicated but WELL documented.    *
 ******************************************************************************/

#include "xpf_internal.h"

extern int	bordercmp();
extern int	memmaskcpy();
#if	STRINGTYPE && BYTEADDRESS
extern int	strnmaskordercmp();
extern int	strnmaskcpy();
#endif	STRINGTYPE && BYTEADDRESS

#define DOPRINT(s)
#define NOPRINT(s)
#define NOWPRINT(s)
#define ERRPRINT(s)	s

/*
 * Returns NULL if pvalue does not exist, the hashentry otherwise.
 * Must specify the ptr to the value (casted appropriately) and the
 * +ve size of the datatype. Can afford to be inefficient (but clean) here.
 */
HashEntry **
xpfGetHashEntry(hashtable, index, pvalue, size)
	HashEntry	**hashtable;
	int		index;
	XPFDataUnit	*pvalue;
	int		size;
{
	HashEntry	**pe;

	if (hashtable == NULL) return NULL;
	pe = &hashtable[index];
	while ((*pe != NULL) && (*(unsigned long *)(*pe)->common.hash.value != *(unsigned long *)pvalue))
		pe = &(*pe)->or;
	return pe;
}

/*
 * Prints out the PAT tree starting at "node" at the given "level". Called
 * initially with the root at level 0. Convention: up is and, down is or.
 */
void
xpfPrintPATTree(node, level)
	PATNode		*node;
	int		level;
{
	int		i = level, j;
	XPFPattern	*p;
	XPFFilter	*f;
	HashEntry	**hashtable;
	PATNode		*n;

	if (node == NULL) {
		while(i-->0) xpfPrintf("  ");
		xpfPrintf("  :NULL\n");
		return;
	}
	xpfPrintPATTree(node->and, level+1);
	while(i-->0) xpfPrintf("  ");

	if (!(node->nodetype & XPF_DAGS_HASHENTRY)) {
		xpfPrintf("  :%x#%x#%x ", node->refcount, node->nodetype, node->info);
		xpfPrintPatternList(&node->common.pattern);
		xpfPrintf("\n");
		p = &node->common.pattern;
		while (p!=NULL) {
			f = p->filterlist;
			while(f!=NULL) {
				if (f->extractor.datatype < 0) {
					hashtable = (HashEntry **)(*(unsigned long *)f->value);
					for (j=0; j<XPF_DAGS_HASHTABLESIZE; j++) {
						n = hashtable[j];
						while(n!=NULL) {
							xpfPrintPATTree(n, level);
							n = n->or;	/* next hash entry */
						}
					}
					goto break_break;
				}
				f = f->next;
			}
			p = p->next;
		}
	break_break:
		xpfPrintPATTree(node->or, level+1);
	}
	else {
		xpfPrintf("  *%x#%x#%x ", node->refcount, node->nodetype, node->info);
		xpfPrintf("<%x>", *(unsigned long *)node->common.hash.value);
		xpfPrintProgramList(node->common.hash.programlist);
		xpfPrintf("\n");
		/* hashentries have no "or" value */
	}
}

/*
 * Orders the filters srcfilter and destfilter so that the order of insertion can be determined. Returns XPF_DAGS_AFTER
 * or XPF_DAGS_BEFORE according to whether srcfilter should be inserted AFTER or BEFORE destfilter. It returns XPF_DAGS_SAME
 * if they are identical and XPF_DAGS_ERROR if there is some problem. It also sets the variable commonprefix
 * (if it exists) to the length of the common prefix between srcfilter and destfilter. The return values determine
 * the "ordering" between the filters AFTER the common prefix has been extracted.
 * REMEMBER: BEFORE IS EARLIER IN DECREASINTG ORDER, AFTER IS LATER IN DECREASING ORDER: MOST SPECIFIC MASK FIRST ARGUMENT.
 *
 * Note that srcfilter is ALWAYS single valued, while destfilter can be multivalued: we must check for that.
 * Also, we cannot split multivalued extractors since we'd have to split each value too, a tremendous overhead.
 */
#if	OVERLAP

#define xpfOrderFilters(srcfilter, destfilter, commonprefix, srcbytesleft, destbytesleft, granularity, ret, end_order)\
{\
	int		srclevel = srcfilter->extractor.selector.level;\
	int		beginsrc = srcfilter->extractor.selector.offset XPF_USER_GETCOUNTUNIT(srcfilter);\
	int		endsrc;\
	int		destlevel = destfilter->extractor.selector.level;\
	int		begindest = destfilter->extractor.selector.offset XPF_USER_GETCOUNTUNIT(destfilter);\
	int		enddest;\
	int		maxprefix;\
	int		size;\
\
	commonprefix = 0;\
	if (srclevel < destlevel) {\
		ret = XPF_DAGS_BEFORE;\
		goto end_order;\
	}\
	else if (srclevel > destlevel) {\
		ret = XPF_DAGS_AFTER;\
		goto end_order;\
	}\
	if (srcfilter->extractor.datatype < 0) {\
		ret = XPF_DAGS_ERROR;\
		goto end_order;\
	}\
	if ( (srcfilter->extractor.datatype == XPF_DATA_NONE) && (destfilter->extractor.datatype == XPF_DATA_NONE) ) {\
		ret = XPF_DAGS_SAME;\
		goto end_order;\
	}\
	else if (srcfilter->extractor.datatype == XPF_DATA_NONE) {\
		ret = XPF_DAGS_AFTER;\
		goto end_order;\
	}\
	else if (destfilter->extractor.datatype == XPF_DATA_NONE) {\
		ret = XPF_DAGS_BEFORE;\
		goto end_order;\
	}\
	/* actually, the above case should not occur: i.e., if destfilter...datatype is XPF_DATA_NONE, then we should skip over it */\
\
	endsrc = beginsrc + xpfSizeFromType(srcfilter->extractor.datatype);\
	enddest= begindest + xpfSizeFromType(destfilter->extractor.datatype);\
	size = endsrc - beginsrc;\
	srcbytesleft = destbytesleft = size;\
\
	/*\
	 * Since headers are of variable lengths, you cannot check if endsrc/enddest can\
	 * overshoot the header size at THIS time (insertion-time): can do ONLY @ SEARCH time.\
	 * Might want to do it for fixed size headers here though.\
	 */\
\
	NOPRINT(xpfPrintf("bs %x es %x bd %x ed %x\n", beginsrc, endsrc, begindest, enddest);)\
	/* check for two cases without overlapp: note that the "end"*s are uppper limit on indices, i.e., valid region = begin to < end */\
	if (beginsrc >= enddest) {\
		ret = XPF_DAGS_AFTER;\
		goto end_order;\
	}\
	else if (begindest >= endsrc) {\
		ret = XPF_DAGS_BEFORE;\
		goto end_order;\
	}\
	else if (beginsrc == begindest) {\
		/* doesn't matter if value is equal to some value or not: we'll hash and ++refcount or insert afresh */\
		if (*(unsigned long *)srcfilter->extractor.mask == *(unsigned long *)destfilter->extractor.mask)\
			ret = XPF_DAGS_SAMEMASK;\
                if ((ret = bordercmp(srcfilter->extractor.mask, destfilter->extractor.mask, size)) == 0) {\
			ret = XPF_DAGS_SAMEMASK;\
			goto end_order;\
		}\
		else if (destfilter->extractor.datatype < 0) {	/* cannot split a multivalued extractor -- too complicated: order arithmetically */\
			ret = (ret == -1) ? XPF_DAGS_AFTER : XPF_DAGS_BEFORE;\
			goto end_order;\
		}\
\
		/* else check for the case where they can have a common prefix: if so, both mask and value must be equal within the prefix */\
		maxprefix = ((endsrc <= enddest) ? endsrc : enddest) - beginsrc;	/* is atmost XPF_DATA_MAXSIZE */\
		ret = XPF_DAGS_SAME;\
		for (commonprefix=0; commonprefix<maxprefix; (commonprefix)++) {\
			if (srcfilter->extractor.mask[commonprefix] > destfilter->extractor.mask[commonprefix]) {\
				ret = XPF_DAGS_BEFORE;\
				break;\
			} else if (srcfilter->extractor.mask[commonprefix] < destfilter->extractor.mask[commonprefix]) {\
				ret = XPF_DAGS_AFTER;\
				break;\
			}\
			else if (srcfilter->value[commonprefix] > destfilter->value[commonprefix]) {\
				ret = XPF_DAGS_BEFORE;\
				break;\
			}\
			else if (srcfilter->value[commonprefix] < destfilter->value[commonprefix]) {\
				ret = XPF_DAGS_AFTER;\
				break;\
			}\
		}\
\
		if (commonprefix % granularity) commonprefix = ((commonprefix) / granularity) * granularity;	/* granularity control */\
		NOPRINT(xpfPrintf("\tprefix %x\n", commonprefix);)\
		srcbytesleft = endsrc - beginsrc - commonprefix;\
		destbytesleft = enddest - begindest - commonprefix;\
		if ((srcbytesleft == 0) && (destbytesleft == 0)) ret = XPF_DAGS_SAME;	/* should never happen though! */\
		goto end_order;\
	}\
	/* order simply based on offset of the beginning of the field selected */\
	else if (beginsrc > begindest) ret = XPF_DAGS_AFTER;\
	else ret = XPF_DAGS_BEFORE;\
end_order:\
	;\
}

#else	OVERLAP

#define xpfOrderFilters(srcfilter, destfilter, commonprefix, srcbytesleft, destbytesleft, granularity, ret, end_order)\
{\
	int		srclevel = srcfilter->extractor.selector.level;\
	int		beginsrc = srcfilter->extractor.selector.offset XPF_USER_GETCOUNTUNIT(srcfilter);\
	int		endsrc;\
	int		destlevel = destfilter->extractor.selector.level;\
	int		begindest = destfilter->extractor.selector.offset XPF_USER_GETCOUNTUNIT(destfilter);\
	int		enddest;\
\
	commonprefix = 0;\
	if (srclevel < destlevel) {\
		ret = XPF_DAGS_BEFORE;\
		goto end_order;\
	}\
	else if (srclevel > destlevel) {\
		ret = XPF_DAGS_AFTER;\
		goto end_order;\
	}\
	if (srcfilter->extractor.datatype < 0) {\
		ret = XPF_DAGS_ERROR;\
		goto end_order;\
	}\
	if ( (srcfilter->extractor.datatype == XPF_DATA_NONE) && (destfilter->extractor.datatype == XPF_DATA_NONE) ) {\
		ret = XPF_DAGS_SAME;\
		goto end_order;\
	}\
	else if (srcfilter->extractor.datatype == XPF_DATA_NONE) {\
		ret = XPF_DAGS_AFTER;\
		goto end_order;\
	}\
	else if (destfilter->extractor.datatype == XPF_DATA_NONE) {\
		ret = XPF_DAGS_BEFORE;\
		goto end_order;\
	}\
	/* actually, the above case should not occur: i.e., if destfilter...datatype is XPF_DATA_NONE, then we should skip over it */\
\
	endsrc = beginsrc + xpfSizeFromType(srcfilter->extractor.datatype);\
	enddest= begindest + xpfSizeFromType(destfilter->extractor.datatype);\
\
	/*\
	 * Since headers are of variable lengths, you cannot check if endsrc/enddest can\
	 * overshoot the header size at THIS time (insertion-time): can do ONLY @ SEARCH time.\
	 * Might want to do it for fixed size headers here though.\
	 */\
\
	NOPRINT(xpfPrintf("bs %x es %x bd %x ed %x\n", beginsrc, endsrc, begindest, enddest);)\
	/* check for two cases without overlapp: note that the "end"*s are uppper limit on indices, i.e., valid region = begin to < end */\
	if (beginsrc >= enddest) {\
		ret = XPF_DAGS_AFTER;\
		goto end_order;\
	}\
	else if (begindest >= endsrc) {\
		ret = XPF_DAGS_BEFORE;\
		goto end_order;\
	}\
	else if (beginsrc == begindest) {\
		/* doesn't matter if value is equal to some value or not: we'll hash and ++refcount or insert afresh */\
		if (*(unsigned long *)srcfilter->extractor.mask == *(unsigned long *)destfilter->extractor.mask)\
			ret = XPF_DAGS_SAMEMASK;\
		else if (*(unsigned long *)srcfilter->extractor.mask > *(unsigned long *)destfilter->extractor.mask)\
			ret = XPF_DAGS_BEFORE;\
		else ret = XPF_DAGS_AFTER;\
		goto end_order;\
	}\
	/* order simply based on offset of the beginning of the field selected */\
	else if (beginsrc > begindest) ret = XPF_DAGS_AFTER;\
	else ret = XPF_DAGS_BEFORE;\
end_order:\
	;\
}

#endif	OVERLAP

/*
 * This is sensible only when the location of patterns is "SRCNEWDESTNEW".
 * This function orders whole patterns at one go --> processes get
 * a higher priority than patterns so that they cannot be merged
 * with each other (kept logically separate). The flags are set in Launch.
 * This is called from Get/Insert/Delete: a user is not allowed to access
 * processes directly: they are deleted indirectly by specifying "alarms".
 */
#define xpfOrderPatterns(srcpattern, destpattern) \
	( ((srcpattern)->flags & XPF_FLAG_PROCESS) ? (((destpattern)->flags & XPF_FLAG_PROCESS) ? XPF_DAGS_SAME : XPF_DAGS_BEFORE) : \
		(((destpattern)->flags & XPF_FLAG_PROCESS) ? XPF_DAGS_AFTER : XPF_DAGS_SAME) )

/*
 * Returns one of DESTPREFIX, SRCPREFIX, SAME, BEFORE or AFTER: used in
 * interface routine SetupSuffix to detect possible cycles in dags.
 * If the tail recursion in the Get/Insert/Delete routines is "removed",
 * they will have a skeleton similar to that of this function. These are
 * two simple patternlists without any hash-tables in their filters's values.
 */
int
xpfOrderPatternLists(src, dest, grain)
	XPFPattern	*src, *dest;
	int		grain;
{
	XPFPattern	*sp, *dp;
	XPFFilter	*sf = NULL, *df = NULL;
	int		ret, commonprefix, srcbytesleft, destbytesleft;

	sp = src;
	dp = dest;
	while ((sp != NULL) && (dp != NULL)) {
		if (sf == NULL) sf = sp->filterlist;
		if (df == NULL) df = dp->filterlist;

		if ( (ret = xpfOrderPatterns(sp, dp)) == XPF_DAGS_AFTER) return XPF_DAGS_AFTER;
		else if (ret == XPF_DAGS_BEFORE) return XPF_DAGS_BEFORE;
		while ((sf != NULL) && (df != NULL)) {
			xpfOrderFilters((sf), (df), (commonprefix), (srcbytesleft), (destbytesleft), (grain), (ret), end_order1);
			if (ret == XPF_DAGS_AFTER) return XPF_DAGS_AFTER;
			else if (ret == XPF_DAGS_BEFORE) return XPF_DAGS_BEFORE;
			else if (ret == XPF_DAGS_SAMEMASK) {
				if (*(unsigned long *)sf->value > *(unsigned long *)df->value) return XPF_DAGS_BEFORE;
				else if (*(unsigned long *)sf->value < *(unsigned long *)df->value) return XPF_DAGS_AFTER;
			}
			else if (ret == XPF_DAGS_ERROR) return XPF_DAGS_ERROR;
			/* else the filters are identical */
			sf = sf->next;
			df = df->next;
		}

		if (sf == NULL) {
			if (df == NULL) {
				sp = sp->next;
				dp = dp->next;
			}
			else {
				if (sp->next == NULL) return XPF_DAGS_SRCPREFIX;
				else {
					sp = sp->next;
				}
			}
		}
		else {
			if (df == NULL) {
				if (dp->next == NULL) return XPF_DAGS_DESTPREFIX;
				else {
					dp = dp->next;
				}
			}
			else ERRPRINT(xpfPrintf("xpdOrderPatternLists: cannot come here!\n"));
		}
	}

	return ( (sp == NULL) ? ((dp == NULL) ? XPF_DAGS_SAME : XPF_DAGS_SRCPREFIX) : ((dp == NULL) ? XPF_DAGS_DESTPREFIX : XPF_DAGS_SAME) );
}

/*
 * Returns a Position that matches the given patternlist exactly.
 * Assumes that if a node is not NULL, then it has atleast one pattern
 * in the patternlist and one filter in the pattern's filterlist.
 * Completely skips over processes since this is for the user-interface only.
 *
 * NOTE: you don't need ptr to ptr to PATNode,pattern,filter as the 3
 * destination parameters, but it becomes easy to compare with the corr.
 * insert/delete routines and double-check ALL 3 routines at the same time.
 */
int
xpfGetPositionFromPatternList(state, srcpattern, srcfilter, proot, destfilter, granularity, location, Pproot, Pdestfilter)
	XPFState	*state;
	XPFPattern	*srcpattern;
	XPFFilter	*srcfilter;	
	PATNode		**proot;
	XPFFilter	**destfilter;
	int		granularity;
	int		location;
	PATNode		***Pproot;
	XPFFilter	***Pdestfilter;
{
	int		ret, commonprefix, srcbytesleft, destbytesleft;
	HashEntry	**pe;
	int		index, size;

	DOPRINT(xpfPrintf("In xpfGetPositionFromPatternList: location = %x\n", location);)
	/*
	 * You should check for the possibility that srcfilter / srcpattern can
	 * become NULL before making ANY recursive call.
	 */
	if ((srcfilter == NULL) || (srcpattern == NULL)) return FALSE;

	if (*proot == NULL) return FALSE;

	/* skip over processes */
	if ((*proot)->nodetype & XPF_DAGS_PROCESS) {
		if ((*proot)->or == NULL) return FALSE;
		else return xpfGetPositionFromPatternList(state,
							srcpattern,
							srcfilter,
							&((*proot)->or),
							&((*proot)->or->common.pattern.filterlist),
							granularity,
							location | XPF_DAGS_DESTISNEW,
							Pproot,
							Pdestfilter);
	}

	if (*destfilter == NULL) {
		NOPRINT(xpfPrintf("\tGoto NEW proot\n");)
		if ((*proot)->and != NULL) {
			state->xpfMatchedNodes[state->xpfMatchedNum ++] = proot;
			return xpfGetPositionFromPatternList(state,
							srcpattern,
							srcfilter,
							&((*proot)->and),
							&((*proot)->and->common.pattern.filterlist),
							granularity,
							location | XPF_DAGS_DESTISNEW,
							Pproot,
							Pdestfilter);
		}
		else return FALSE;
	}

	DOPRINT(xpfAssert((*destfilter != NULL) && (*proot != NULL)));

	/*
	 * When we are looking for filters with type=XPF_DATA_NONE (i.e., match all), we have to make sure that they
	 * have the lowest priority, i.e., they are AFTER everything and alone in any filterlist.
	 */

	if ( (location == XPF_DAGS_SRCOLDDESTOLD) ||
	     ((location == XPF_DAGS_SRCNEWDESTOLD) && (srcpattern->header.headertype & XPF_HEADER_CURRENT)) ||
	     ((location == XPF_DAGS_SRCOLDDESTNEW) && ((*proot)->common.pattern.header.headertype & XPF_HEADER_CURRENT)) ||
	     ((location == XPF_DAGS_SRCNEWDESTNEW) && (srcpattern->header.headertype & XPF_HEADER_CURRENT) && ((*proot)->common.pattern.header.headertype & XPF_HEADER_CURRENT)) )
	{
		/*
		 * CASE 1: Both within the same header as before.
		 */
	case1G:
		if (location == XPF_DAGS_SRCNEWDESTNEW) {
			if ((ret = xpfOrderPatterns(srcpattern, &(*proot)->common.pattern)) == XPF_DAGS_AFTER) goto case2G;
			else if (ret == XPF_DAGS_BEFORE) goto case3G;
		}
		NOPRINT(xpfPrintf("\tOLD headers\n");)
		if (srcfilter->extractor.datatype == XPF_DATA_NONE) {	/* look for identical match: if not, goto the next destfilter */
			if (srcfilter->next == NULL) {
				if ((*destfilter)->extractor.datatype == XPF_DATA_NONE) {
					state->xpfMatchedNodes[state->xpfMatchedNum ++] = proot;
					*Pproot = proot;
					*Pdestfilter = destfilter;
					return TRUE;
				}
				/* else order the filters: you will get srcfilter AFTER destfilter anyway with 0 common-prefix */
			}
			else {	/* compare srcfilter->next with destfilter: if identical skip TWICE in src; skip once in dest (in anycase) */
				if (srcfilter->next->extractor.datatype == XPF_DATA_NONE)	/* just ignore one "*" in **<str> */
					return xpfGetPositionFromPatternList(state,
									srcpattern,
									srcfilter->next,
									proot,
									destfilter,
									granularity,
									XPF_DAGS_SRCOLDDESTOLD,
									Pproot,
									Pdestfilter);

				xpfOrderFilters((srcfilter->next), (*destfilter), (commonprefix), (srcbytesleft), (destbytesleft), (granularity), (ret), end_order2);
				if (ret != XPF_DAGS_SAME)
					if (ret == XPF_DAGS_ERROR) return FALSE;
					else return xpfGetPositionFromPatternList(state,
										srcpattern,
										srcfilter,
										proot,
										&(*destfilter)->next,
										granularity,
										XPF_DAGS_SRCOLDDESTOLD,
										Pproot,
										Pdestfilter);
					/* just as if srcfilter matched destfilter but we don't advance srcfilter */
				else { /* This is just as if the ordering below produced XPF_DAGS_SAME without the need to copy anything */
					if (srcfilter->next->next == NULL)
						if (srcpattern->next == NULL) {
							state->xpfMatchedNodes[state->xpfMatchedNum ++] = proot;
							*Pproot = proot;
							*Pdestfilter = destfilter;
							return TRUE;
						}
						else return xpfGetPositionFromPatternList(state,
											srcpattern->next,
											srcpattern->next->filterlist,
											proot,
											&(*destfilter)->next,
											granularity,
											XPF_DAGS_SRCNEWDESTOLD,
											Pproot,
											Pdestfilter);
					else return xpfGetPositionFromPatternList(state,
										srcpattern,
										srcfilter->next->next,
										proot,
										&(*destfilter)->next,
										granularity,
										XPF_DAGS_SRCOLDDESTOLD,
										Pproot,
										Pdestfilter);
				}
			}
		}

		if ((*destfilter)->extractor.datatype == XPF_DATA_NONE) {	/* lowest priority: if you order srcfilter and *destfilter, you'll be in trouble! */
			srcpattern->header.headertype |= XPF_HEADER_CURRENT;	/* since this is like an "after" */
			return xpfGetPositionFromPatternList(state,
							srcpattern,
							srcfilter,
							proot,
							&(*destfilter)->next,
							granularity,
							XPF_DAGS_SRCOLDDESTOLD,
							Pproot,
							Pdestfilter);
		}

		xpfOrderFilters((srcfilter), (*destfilter), (commonprefix), (srcbytesleft), (destbytesleft), (granularity), (ret), end_order3);
		if(ret == XPF_DAGS_SAMEMASK) {
			DOPRINT(xpfPrintf("\t\tXPF_DAGS_SAMEMASK see hash\n");)

			size = xpfSizeFromType(srcfilter->extractor.datatype);
			index = xpfHashIndex(srcfilter->value);

			if ((*destfilter)->extractor.datatype > 0) {
				if (*(unsigned long *)srcfilter->value == *(unsigned long *)(*destfilter)->value) {
					/* same as XPF_DAGS_SAME with common prefix = size */
					commonprefix = size;
					srcbytesleft = 0;
					destbytesleft = 0;
					goto identicalG;
				}
				else if (*(unsigned long *)srcfilter->value > *(unsigned long *)(*destfilter)->value) {
					commonprefix = 0;
					ret = XPF_DAGS_BEFORE;
					goto notidenticalG;
				}
				else {
					commonprefix = 0;
					ret = XPF_DAGS_AFTER;
					goto notidenticalG;
				}
			}
			else {
				pe = xpfGetHashEntry( (HashEntry **)*(unsigned long *)((*destfilter)->value), index, srcfilter->value, size);
				if (*pe == NULL) {
					commonprefix = 0;
					ret = XPF_DAGS_AFTER;
					goto notidenticalG;	/* cannot say proot has matched ... */
					/* cannot be BEFORE else wouldn't have come here: if at all it can occur if same *destfilter REPEATS later */
				}
				else {	/* The assertion proot != NULL still holds but *pe->and can be NULL: must check here */
					state->xpfMatchedNodes[state->xpfMatchedNum ++] = proot;	/* this pattern has got to be the last pattern */
					state->xpfMatchedNodes[state->xpfMatchedNum ++] = pe;
					if (srcfilter->next == NULL) {
						if (srcpattern->next != NULL) {
							if ((*pe)->and == NULL) return FALSE;
							else return xpfGetPositionFromPatternList(state,
												srcpattern->next,
												srcpattern->next->filterlist,
												&(*pe)->and,
												&(*pe)->and->common.pattern.filterlist,
												granularity,
												XPF_DAGS_SRCNEWDESTNEW,
												Pproot,
												Pdestfilter);
						}
						else {
							*Pproot = pe;
							*Pdestfilter = destfilter;
							return TRUE;
						}
					}
					else {
						if ((*pe)->and == NULL) return FALSE;
						else return xpfGetPositionFromPatternList(state,
											srcpattern,
											srcfilter->next,
											&(*pe)->and,
											&(*pe)->and->common.pattern.filterlist,
											granularity,
											XPF_DAGS_SRCOLDDESTNEW,
											Pproot,
											Pdestfilter);
					}
				}
			}
		}
		else if (ret == XPF_DAGS_SAME) {
		identicalG:
			NOPRINT(xpfPrintf("\t\tXPF_DAGS_SAME advance\n"));
			if (srcfilter->next == NULL) {
				if (srcpattern->next != NULL)
					return xpfGetPositionFromPatternList(state,
									srcpattern->next,
									srcpattern->next->filterlist,
									proot,
									&(*destfilter)->next,
									granularity,
									XPF_DAGS_SRCNEWDESTOLD,
									Pproot,
									Pdestfilter);
				else {
					state->xpfMatchedNodes[state->xpfMatchedNum ++] = proot;
					*Pproot = proot;
					*Pdestfilter = destfilter;
					return TRUE;
				}
			}
			else return xpfGetPositionFromPatternList(state,
								srcpattern,
								srcfilter->next,
								proot,
								&(*destfilter)->next,
								granularity,
								XPF_DAGS_SRCOLDDESTOLD,
								Pproot,
								Pdestfilter);
		}
		else if (ret == XPF_DAGS_ERROR) return FALSE;
		else {
		notidenticalG:
			if (commonprefix > 0) {	/* handled differently even if XPF_DAGS_AFTER/BEFORE is known since keys themselves can change */
				NOPRINT(xpfPrintf("\t\tCOMPREFIX: adjust\n");)

				/* cannot find exact match: user should give what (s)he inserted not half of it: get does not alter the tree! */
				if (destbytesleft > 0) return FALSE;
				destfilter = &(*destfilter)->next;

				if (srcbytesleft > 0) {
					xpfMemcpy(srcfilter->value, srcfilter->value + commonprefix, srcbytesleft);
					xpfMemcpy(srcfilter->extractor.mask, srcfilter->extractor.mask + commonprefix, srcbytesleft);
					xpfBzero((char *)(srcfilter->value + srcbytesleft), commonprefix);
					xpfBzero((char *)(srcfilter->extractor.mask + srcbytesleft), commonprefix);
					srcfilter->extractor.datatype = srcbytesleft;
					srcfilter->extractor.selector.offset =
						srcfilter->extractor.selector.offset XPF_USER_GETCOUNTUNIT(srcfilter) + commonprefix;
#if	COUNTUNIT
					srcfilter->extractor.selector.countunit = 1;
#endif	COUNTUNIT
				}
				else srcfilter = srcfilter->next;

				if (srcfilter == NULL) {
					if (srcpattern->next != NULL)
						return xpfGetPositionFromPatternList(state,
										srcpattern->next,
										srcpattern->next->filterlist,
										proot,
										destfilter,
										granularity,
										XPF_DAGS_SRCNEWDESTOLD,
										Pproot,
										Pdestfilter);
					else {
						state->xpfMatchedNodes[state->xpfMatchedNum ++] = proot;
						*Pproot = proot;
						*Pdestfilter = destfilter;
						return TRUE;
					}
				}
				else return xpfGetPositionFromPatternList(state,
									srcpattern,
									srcfilter,
									proot,
									destfilter,
									granularity,
									XPF_DAGS_SRCOLDDESTOLD,
									Pproot,
									Pdestfilter);
			}
			else if (ret == XPF_DAGS_AFTER) {
				/*
				 * Very similar to CASE 2: I have to look for source after the destination.
				 */
			case2G:
				NOPRINT(xpfPrintf("\t\tXPF_DAGS_AFTER: advance\n");)

				srcpattern->header.headertype |= XPF_HEADER_CURRENT;	/* src+dest were in same hdr => src's header shouldn't be extracted */
				if ((*proot)->or != NULL)
					return xpfGetPositionFromPatternList(state,
									srcpattern,
									srcfilter,
									&(*proot)->or,
									&((*proot)->or->common.pattern.filterlist),
									granularity,
									XPF_DAGS_SRCOLDDESTNEW,
									Pproot,
									Pdestfilter);
				else return FALSE;

			}
			else {	/* neither SAME nor AFTER: => not found */
			case3G:
				return FALSE;
			}
		}
	}
	else if ( (location == XPF_DAGS_SRCNEWDESTOLD) ||
		  ((location == XPF_DAGS_SRCNEWDESTNEW) && ((*proot)->common.pattern.header.headertype & XPF_HEADER_CURRENT)) )
	{
		/*
		 * CASE 2: srcfilter is in a new header, destfilter is still in the old header.
		 * Current protocol always has a higher priority over another protocol.
		 * Control comes here also when srcpattern (as a whole) is AFTER proot.pattern.
		 */
		NOPRINT(xpfPrintf("\tSRC=new, DEST=old\n");)
		if ((*proot)->or != NULL)
			return xpfGetPositionFromPatternList(state,
							srcpattern,
							srcfilter,
							&(*proot)->or,
							&((*proot)->or->common.pattern.filterlist),
							granularity,
							XPF_DAGS_SRCNEWDESTNEW,
							Pproot,
							Pdestfilter);
		else return FALSE;
	}
	else if ( (location == XPF_DAGS_SRCOLDDESTNEW) ||
		  ((location == XPF_DAGS_SRCNEWDESTNEW) && (srcpattern->header.headertype & XPF_HEADER_CURRENT)) )
	{
		/*
		 * CASE 3: srcfilter is not in a new header, but destfilter is.
		 * so srcpattern cannot be found in the destination in anyway.
		 * Control comes here also when srcpattern (as a whole) is BEFORE proot.pattern.
		 */
		NOPRINT(xpfPrintf("\tSRC=old, DEST=new\n");)
		return FALSE;
	}
	else {
		/*
		 * CASE 4: Both are in new headers. The headers must match exactly. Should
		 */
		NOPRINT(xpfPrintf("\tNEW headers\n");)
		if (srcpattern->header.headertype != (*proot)->common.pattern.header.headertype)
			return FALSE;
#if	VARHEADER
		if (srcpattern->header.headertype & XPF_HEADER_VARIABLE) {
			if (bcmp(srcpattern->header.size.variable, (*proot)->common.pattern.header.size.variable, sizeof(XPFVariableHeader)) != 0) {
				/* Faster than checking individual fields */
				ERRPRINT(xpfPrintf("Attempt to change header of pattern!\n"));
				return FALSE;
			}
		}
		else
#endif	VARHEADER
		if (srcpattern->header.size.fixed != (*proot)->common.pattern.header.size.fixed)
			return FALSE;

		goto case1G;
	}
}

/*
 * Some mutually recursive destructor functions.
 */

void xpfDestroyPATNode();	/* forward */

/*
 * Completely ruins the subtree rooted at node.
 * Called only when both the children of the initial "pnode" argument
 * are NULL, the filterlist in the node is NULL, its reference count
 * is 0 or both the children of *pnode look at headers of DIFFERENT
 * protocols (a serious error: IP deleted w/o TCP being deleted) when
 * neither can be inserted into the other.
 *
 * NOTE: the refcount of the *pnode must be decremented before call.
 * This deletes the tree only if the node is NOT being used by anyone.
 * This (and functions called from it) do(es) not examine the refcounts
 * of the individual filterlist elements since this function is called
 * when the pnode itself "should not" exist in the PATTree.
 */
void
xpfDestroyDeadPATTree(pnode
#if	MERGEINFO
			,deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
			,deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
			)
	PATNode	**pnode;
#if	MERGEINFO
	XPFInfoList	**deletedinfo;
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
	XPFMessageList	**deletedmessagelist;
#endif	DELAYEDMSGS && STOREMSGS
{
	if (*pnode == NULL) return;
	if ((*pnode)->refcount > 0) return;	/* must decrement it before call if reqd. */
	NOPRINT(xpfPrintf("Destroy subtree @ %x\n", *pnode);)
	if ((*pnode)->and != NULL) {
		(*pnode)->and->refcount --;
		xpfDestroyDeadPATTree(&(*pnode)->and
#if	MERGEINFO
					,deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
					,deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
					);
	}
	if ((*pnode)->or != NULL) {
		(*pnode)->or->refcount --;
		xpfDestroyDeadPATTree(&(*pnode)->or
#if	MERGEINFO
					, deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
					, deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
					);
	}
	xpfDestroyPATNode(pnode
#if	MERGEINFO
			, deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
			, deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
			);
}

/*
 * Completely ruins the subtree rooted at node and returns all messages/infos.
 * Moreover, this funtion FIRST decrements the reference count and THEN
 * deletes the node => it is guaranteed to remove all nodes of the DAG NOW.
 */
void
xpfDestroyLivePATTree(pnode
#if	MERGEINFO
			, deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
			, deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
			)
	PATNode	**pnode;
#if	MERGEINFO
	XPFInfoList	**deletedinfo;
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
	XPFMessageList	**deletedmessagelist;
#endif	DELAYEDMSGS && STOREMSGS
{
	if (*pnode == NULL) return;
	NOPRINT(xpfPrintf("Destroy subtree @ %x\n", *pnode);)
	if ((*pnode)->and != NULL) {
		xpfDestroyLivePATTree(&(*pnode)->and
#if	MERGEINFO
					, deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
					, deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
					);
	}
	if ((*pnode)->or != NULL) {
		xpfDestroyLivePATTree(&(*pnode)->or
#if	MERGEINFO
					, deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
					, deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
					);
	}
	if (--(*pnode)->refcount > 0) return;	/* another path will get you! */
	xpfDestroyPATNode(pnode
#if	MERGEINFO
			, deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
			, deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
			);
}

/*
 * Completely ruins the given filterlist knowing that some of the filters can
 * have hash-tables as their values (compare with xpf_user.h/xpfDestroyFilterList).
 */
void
xpfDestroyDAGSFilterList(statlist
#if	MERGEINFO
			, deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
			, deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
			)
	XPFFilter	*statlist;
#if	MERGEINFO
	XPFInfoList	**deletedinfo;
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
	XPFMessageList	**deletedmessagelist;
#endif	DELAYEDMSGS && STOREMSGS
{
	XPFFilter	*s, *ts;

	s = statlist;
	while (s != NULL) {
		if (s->extractor.datatype < 0) {
			/* no need to use xpfDeleteHashEntry, etc., since this is global destruction! */
			int		i;
			HashEntry	*te, *e, **hashtable = (HashEntry **)(*(unsigned long *)s->value);

			for (i=0; i<XPF_DAGS_HASHTABLESIZE; i++) {
				e = hashtable[i];

				while (e!= NULL) {
					if (e->and != NULL) {
						if(--e->and->refcount <= 0) {	/* this node cannot have any references outside "e" */
							xpfDestroyDeadPATTree(&e->and
#if	MERGEINFO
										,deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
										,deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
										);
						}
					}
					te = e;
					e = e->or;
					xpfDestroyHashEntry(te);
				}
			}
			xpfFree(hashtable, sizeof(HashEntry *) * XPF_DAGS_HASHTABLESIZE);
		}
		ts = s;
		s = s->next;
		xpfFree((char *)ts, sizeof(XPFFilter));
	}
}

/*
 * Completely ruins the data-structures at the given node and frees it.
 * Does not check for reference counts: must be changed outside if reqd.
 * The pnode will be a hash-entry only if called from within Destroy-
 * DAGSFilterList().
 */
void
xpfDestroyPATNode(pnode
#if	MERGEINFO
		, deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
		, deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
		)
	PATNode		**pnode;
#if	MERGEINFO
	XPFInfoList	**deletedinfo;
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
	XPFMessageList	**deletedmessagelist;
#endif	DELAYEDMSGS && STOREMSGS
{
#if	DELAYEDMSGS && STOREMSGS
	XPFMessageList  **m;
#endif	DELAYEDMSGS && STOREMSGS
#if	MERGEINFO
	XPFInfoList     **l;
#endif	MERGEINFO
	PATNode		*t;

	if (*pnode != NULL) {
#if	MERGEINFO
		l = &(*pnode)->info;
		while(*l != NULL) l = &(*l)->next;
		*l = *(deletedinfo);
		*(deletedinfo) = (*pnode)->info;
		(*pnode)->info = XPF_INFO_NULL;
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
		m = &(*pnode)->messagelist;
		while(*m != NULL) m = &(*m)->next;
		*m = *(deletedmessagelist);
		*(deletedmessagelist) = (*pnode)->messagelist;
		(*pnode)->messagelist = NULL;
#endif	DELAYEDMSGS && STOREMSGS

		if ((*pnode)->nodetype & XPF_DAGS_HASHENTRY) {
			xpfDestroyProgramList((*pnode)->common.hash.programlist);
			t = *pnode;
			(*pnode) = (*pnode)->or;
			xpfFree((char *)t, sizeof(PATNode));
		}
		else {
			xpfDestroyDAGSFilterList((*pnode)->common.pattern.filterlist
#if	MERGEINFO
						, deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
						, deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
						);
			xpfDestroyProgramList(((*pnode)->common.pattern.programlist));
			xpfFree((char *)(*pnode), sizeof(PATNode));
			*pnode = NULL;
		}
	}
}

/*
 * The following three routines are needed to test if the user's data-structures are in valid space
 * The only routine used from outside here is: xpfIsValidRegion(addr, len) returning TRUE/FALSE.
 */

int
xpfValidFilterList(f)
	XPFFilter	*f;
{
	while(f!=NULL) {
		if (!xpfIsValidRegion(f, sizeof(XPFFilter))) return FALSE;
		f = f->next;
	}
	return TRUE;
}

int
xpfValidProgramList(m)
	XPFFilter	*m;
{
	while(m!=NULL) {
		if (!xpfIsValidRegion(m, sizeof(XPFProgram))) return FALSE;
		m = m->next;
	}
	return TRUE;
}

int
xpfValidPatternList(p)
	XPFPattern	*p;
{
	while (p!=NULL) {
		if (!xpfIsValidRegion(p, sizeof(XPFPattern))) return FALSE;
#if	VARHEADER
		if (p->header.headertype & XPF_HEADER_VARIABLE)
			if (!xpfIsValidRegion(p->header.size.variable, sizeof(XPFVariableHeader))) return FALSE;
#endif	VARHEADER
		if (!xpfValidFilterList(p->filterlist)) return FALSE;
		if (!xpfValidProgramList(p->programlist)) return FALSE;
		p = p->next;
	}
	return TRUE;
}

/*
 * Inserts the given filters with the corr. headers beginning at
 * *proot, after splicing the filters appropriately and creating new
 * PAT-nodes to hold the patterns wherever required. The value in
 * each filters is assumed to be masked with ITS OWN mask before
 * calling this function (mi & vi = vi). During splicing, the program
 * part of the pattern goes into the latter half (the Y branch).
 *
 * Parameters:
 * -----------
 * srcpattern: the patternlist that has to be inserted into the DAG, 
 *	initialliy set to the parameter passed to xControl.
 * srcfilter: current element pointer in patternlist->filterlist being inserted,
 *	initially equal to the filterlist of the patternlist of xControl.
 * proot: place from where we need to begin the search for the point
 *	of insertion, initially pointer to the root of PAT-DAG.
 * destfilter: current element pointer in (*proot)->common.pattern.filterlist being
 *	looked at, initially equal to the pointer to the filterlist of proot.
 * granularity: the indivisible unit in bytes for commmon-prefix matching.
 * location: the relative position of srcfilter in srcpattern and destfilter in
 *	proot's pattern (tells us whether we should compare the headers in the
 *	patternlists or whether we can skip that since we are "within" the lists).
 * flagmask: the mask to be used on the source flag before copying it into the
 *	destination: if called from an interface routine, it is XPF_FLAG_USER
 *	and if called internally (from Launch), then it si XPF_FLAG_DAGS.
 *
 * Return Values:
 * --------------
 * The node at which the last part of the srcpattern was inserted, NULL if error.
 * Note: the function is tail-recursive and hence can be made iterative very
 * easily. However, we don't bother about efficiency at THIS stage!
 *
 * Caution:
 * --------
 * 1. The structure (memory usage) of the input srcpattern is NOT changed by this
 *    function. However, the CONTENTS of its elements might get altered. XPF
 *    makes its own copies of data passed on by the higher level protocols. So,
 *    protocols calling this protocol must delete what they put into the patterns.
 *    Protocols can keep a record of the patternlist-s they allocated to pass-down
 *    the xControl operation, and deallocate them when the operation returns (i.e.,
 *    the contents of the patternlist-s must be considered unusable thereafter).
 *    Larry said OK to the copying part. I don't know about the destroying part.
 *
 * 2. One important invariant which has to be maintained while splicing is that
 *    if a node n is not NULL, we are guaranteed to have atleast one non-null
 *    element in its patternlist and this element should at least one element in its
 *    filterlist. The READJUST macro and the checks for
 *    (destfilter == proot->pattern->filterlist) are used for THIS purpose.
 *
 * 3. Consider combining two patternlists/nodes if possible instead of just splicing:
 *    This should be done by a merge procedure called every once in a while to
 *    merge sequences: a->and->or = a->or (concat filter/program lists of a w/ a->and);
 *    and a->or->and = a->and (concat filter/program lists of a w/ a->or). We can
 *    merge irrespective of header if the or's (case1) and and's (case2) are
 *    NULLs, otherwise, we can merge only the same header guys (headertype = 0).
 *    This saves space for the "nodes" only. Also, time saved is not significant:
 *    so you might postpone it until there is nothing else to do.
 *
 * 4. Invariant: The proot is always one with nodetype & XPF_DAGS_HASHENTRY == 0.
 *    The return value can however be a XPF_DAGS_HASHENTRY type PATNode too.
 */

PATNode *
xpfInsertPatternList(srcpattern, srcfilter, proot, destfilter, granularity, location, flagmask)
	XPFPattern	*srcpattern;
	XPFFilter	*srcfilter;	
	PATNode		**proot;
	XPFFilter	**destfilter;
	int		granularity;
	int		location;
	int		flagmask;
{
	/* Doubt about preserving proper currency information */
	XPFPattern	*k;
	XPFProgram	*d, **pd;
	XPFFilter	*s, **ps;
	PATNode		**prevn, *n, *m;
	int		ret, commonprefix, srcbytesleft, destbytesleft;
	PATNode		**retval;
	XPFDataUnit	pvalue[XPF_DATA_MAXSIZE];
	HashEntry	**pe;
	HashEntry	*e;
	int		size, index;	/* for hash-table search */

	DOPRINT(xpfPrintf("Insert: srcpattern %x srcfilter %x proot %x\n", srcpattern, srcfilter, *proot);)

	/*
	 * You should check for the possibility that srcfilter / srcpattern can
	 * become NULL before making ANY recursive call.
	 */
	if (srcfilter == NULL || srcpattern == NULL) return *proot;

	if (*proot == NULL) {
		if (srcfilter->extractor.datatype == XPF_DATA_NONE) {
			/* skip over the *filter until you reach a non* filter or terminate => auto skip ** */
			if (srcfilter->next != NULL)
				return xpfInsertPatternList(srcpattern,
								srcfilter->next,
								proot,
								NULL /*destfilter */,
								granularity,
								location,
								flagmask);
			else if (srcpattern->next != NULL)
				return xpfInsertPatternList(srcpattern->next,
								srcpattern->next->filterlist,
								proot,
								NULL /* destfilter */,
								granularity,
								location,
								flagmask);
			/* else insert the XPF_DATA_NONE filter into the list anyway: this proot now matched everything */
		}
	
		/*
		 * Copy the whole srcpattern/srcfilter into *proot (one pattern/node).
		 */
		k = srcpattern;
		prevn = NULL;
		while (k != NULL) {
			xpfCreatePATNode(n);
			if (prevn == NULL) {
				*proot = n;	/* first time */
				prevn = proot;
			}
			else {
				(*prevn)->and = n;
				prevn = &(*prevn)->and;
			}
			n->refcount ++;
			n->common.pattern.flags = k->flags & flagmask;
			n->nodetype |= (k->flags & flagmask & XPF_DAGS_PROCESS);
			n->common.pattern.alarm = k->alarm;
			n->common.pattern.processnum = (flagmask == XPF_FLAG_USER) ? 0 : k->processnum;
#if	VARHEADER
			n->common.pattern.header.size.variable = NULL;
#endif	VARHEADER
			COPY_HEADER(&(n->common.pattern.header), &(k->header));
			ps = &n->common.pattern.filterlist;
			if (srcfilter == NULL) srcfilter = k->filterlist;	/* copy from current srcfilter the 1st time */
			s = srcfilter;
			while (s != NULL) {
				*ps = (XPFFilter *)xpfMalloc(sizeof(XPFFilter));
				xpfMemcpy(&((*ps)->extractor), &(s->extractor), sizeof(XPFExtractor));
				(*ps)->extractor.datatype = xpfAbs(s->extractor.datatype);
				xpfMemcpy((*ps)->value, s->value, XPF_DATA_MAXSIZE);
				(*ps)->refcount = 1 /* xpfAbs(s->refcount)*/ ;
				ps = &(*ps)->next;
				s = s->next;
			}
			*ps = NULL;

			pd = &n->common.pattern.programlist;
			d = k->programlist;
			while (d != NULL) {
				*pd = (XPFProgram *)xpfMalloc(sizeof(XPFProgram));
				xpfMemcpy(&((*pd)->extractor), &(d->extractor), sizeof(XPFExtractor));
				(*pd)->extractor.datatype = xpfAbs(d->extractor.datatype);
				(*pd)->refcount = 1 /* xpfAbs(d->refcount) */;
				pd = &(*pd)->next;
				d = d->next;
			}
			*pd = NULL;

			k = k->next;
			srcfilter = NULL;
		}
		DOPRINT(xpfPrintf("xpfInsertPatternList: copied all patternlists for null proot\n");)

		(*proot)->or = (destfilter == NULL) ? NULL : (PATNode *)destfilter;	/* shady way to pass OR info */
		/* guaranteed to have only one pattern = with only 1 header since programs are there for 1 protocol only: very shady way to do things */
		NOPRINT(xpfPrintf("\tNULL-proot: copied all\n");)
		return *prevn;
	}
	ERRPRINT(xpfAssert( (*proot != NULL) && ((*proot)->common.pattern.filterlist != NULL) ));	/* read READJUST_.. macro */

	if (*destfilter == NULL) {
		NOPRINT(xpfPrintf("\tGoto NEW proot\n");)
		/* srcpattern->header.headertype |= XPF_HEADER_CURRENT; ********* can't do it here! 26/3/94 */
		if ((*proot)->and == NULL) return xpfInsertPatternList(srcpattern,
									srcfilter,
									&((*proot)->and),
									NULL,
									NULL,
									granularity,
									location | XPF_DAGS_DESTISNEW,
									flagmask);
		else return xpfInsertPatternList(srcpattern,
						srcfilter,
						&((*proot)->and),
						&((*proot)->and->common.pattern.filterlist),
						granularity,
						location | XPF_DAGS_DESTISNEW,
						flagmask);
	}

	DOPRINT(xpfAssert((*destfilter != NULL) && (*proot != NULL)));

	/*
	 * When we are inserting filters with type=XPF_DATA_NONE (i.e., match all), we have to make sure that they
	 * have the lowest priority, and that if a pattern has a filter with type=XPF_DATA_NONE, it cannot have
	 * any other elements in the filterlist, i.e., "*"=match-all either occurs alone in the set of
	 * filterlist-s, or never occurs at all. This is required to implement the notion of a "not".
	 * Do not forget to increment reference counts of destfilters which match * or the filter after *.
	 */

	if ( (location == XPF_DAGS_SRCOLDDESTOLD) ||
	     ((location == XPF_DAGS_SRCNEWDESTOLD) && (srcpattern->header.headertype & XPF_HEADER_CURRENT)) ||
	     ((location == XPF_DAGS_SRCOLDDESTNEW) && ((*proot)->common.pattern.header.headertype & XPF_HEADER_CURRENT)) ||
	     ((location == XPF_DAGS_SRCNEWDESTNEW) && (srcpattern->header.headertype & XPF_HEADER_CURRENT) && ((*proot)->common.pattern.header.headertype & XPF_HEADER_CURRENT)) )
	{
		/*
		 * CASE 1: Both within the same header as before.
		 */
	case1I:
		if (location == XPF_DAGS_SRCNEWDESTNEW) {
			if ((ret = xpfOrderPatterns(srcpattern, &(*proot)->common.pattern)) == XPF_DAGS_AFTER) goto case2I;
			else if (ret == XPF_DAGS_BEFORE) goto case3I;
		}

		NOPRINT(xpfPrintf("\tOLD headers\n");)
		if (srcfilter->extractor.datatype == XPF_DATA_NONE) {	/* look for identical match: if not, goto the next destfilter */
			if (srcfilter->next == NULL) {
				if ((*destfilter)->extractor.datatype == XPF_DATA_NONE) {
					(*destfilter)->refcount ++;
					return *proot;
				}
				/* else order the filters: you will get srcfilter AFTER destfilter anyway with 0 common-prefix */
			}
			else {	/* compare srcfilter->next with destfilter: if identical skip TWICE in src; skip once in dest (in anycase) */
				if (srcfilter->next->extractor.datatype == XPF_DATA_NONE)	/* just ignore one "*" in **<str> */
					return xpfInsertPatternList(srcpattern,
									srcfilter->next,
									proot,
									destfilter,
									granularity,
									XPF_DAGS_SRCOLDDESTOLD,
									flagmask);
				xpfOrderFilters((srcfilter->next), (*destfilter), (commonprefix), (srcbytesleft), (destbytesleft), (granularity), (ret), end_order4);
				if (ret != XPF_DAGS_SAME) {
					if (ret == XPF_DAGS_ERROR) return NULL;
					else return xpfInsertPatternList(srcpattern,
									srcfilter,
									proot,
									&(*destfilter)->next,
									granularity,
									XPF_DAGS_SRCOLDDESTOLD,
									flagmask);
					/* just as if srcfilter matched destfilter but we don't advance srcfilter */
				}
				else { /* This is just as if the ordering below produced XPF_DAGS_SAME without the need to copy anything */
					(*destfilter)->refcount ++;
					if (srcfilter->next->next == NULL)
						if (srcpattern->next == NULL) return *proot;
						else return xpfInsertPatternList(srcpattern->next,
										srcpattern->next->filterlist,
										proot,
										&(*destfilter)->next,
										granularity,
										XPF_DAGS_SRCNEWDESTOLD,
										flagmask);
					else return xpfInsertPatternList(srcpattern,
									srcfilter->next->next,
									proot,
									&(*destfilter)->next,
									granularity,
									XPF_DAGS_SRCOLDDESTOLD,
									flagmask);
				}
			}
		}

		if ((*destfilter)->extractor.datatype == XPF_DATA_NONE) {	/* lowest priority: cannot order srcfilter and *destfilter! */
			srcpattern->header.headertype |= XPF_HEADER_CURRENT;	/* since this is like an "after" */
			return xpfInsertPatternList(srcpattern,
							srcfilter,
							proot,
							&(*destfilter)->next,
							granularity,
							XPF_DAGS_SRCOLDDESTOLD,
							flagmask);
		}

		xpfOrderFilters((srcfilter), (*destfilter), (commonprefix), (srcbytesleft), (destbytesleft), (granularity), (ret), end_order5);
		if(ret == XPF_DAGS_SAMEMASK) {
			DOPRINT(xpfPrintf("\t\tXPF_DAGS_SAMEMASK see hash\n");)

			size = xpfSizeFromType(srcfilter->extractor.datatype);
			index = xpfHashIndex(srcfilter->value);

			if ((*destfilter)->extractor.datatype > 0) {
				if (*(unsigned long *)srcfilter->value == *(unsigned long *)(*destfilter)->value) {
					/* same as XPF_DAGS_SAME with common prefix = size and both bytes left = 0 */
					ret = XPF_DAGS_SAME;
					commonprefix = size;
					srcbytesleft = 0;
					destbytesleft = 0;
					goto identicalI;	/* don't build hash-table unnecessarily! */
				}
				else {	/* copy the current value into the hash table in an entry, create a hash table and insert new value as ANOTHER entry */
					int	dindex = xpfHashIndex((*destfilter)->value);

					xpfCreateHashTable((*destfilter), pvalue);
					xpfCreateHashEntry(e, pvalue, size);
					e->nodetype |= (*proot)->nodetype & XPF_DAGS_PROCESS;
					e->refcount = (*destfilter)->refcount;	/* not 1 */
					(*destfilter)->extractor.datatype = -(*destfilter)->extractor.datatype;
					xpfInsertHashEntry((HashEntry **)*(unsigned long *)((*destfilter)->value), e, dindex);

					/* Now must determine what e->and should point to and what should be in its programlist */
					if ((*destfilter)->next == NULL) {	/* this is at the end already */
						e->common.hash.processnum = (*proot)->common.pattern.processnum;
						e->common.hash.alarm = (*proot)->common.pattern.alarm;
						e->common.hash.programlist = (*proot)->common.pattern.programlist;
						(*proot)->common.pattern.programlist = NULL;
#if	DELAYEDMSGS && STOREMSGS
						e->messagelist = (*proot)->messagelist;
						(*proot)->messagelist = NULL;
#endif	DELAYEDMSGS && STOREMSGS
						e->info = (*proot)->info;
						(*proot)->info = XPF_INFO_NULL;

						e->and = (*proot)->and;
						(*proot)->and = NULL;
					}
					else {	/* there is something: not at the end of filterlist: no need to copy anything else */
						xpfCreatePATNode(n);
						n->nodetype |= (*proot)->nodetype & XPF_DAGS_PROCESS;
						n->common.pattern.header.headertype /*|*/= XPF_HEADER_CURRENT;
						n->common.pattern.filterlist = (*destfilter)->next;
						(*destfilter)->next = NULL;

						n->common.pattern.programlist = (*proot)->common.pattern.programlist;
						(*proot)->common.pattern.programlist = NULL;

						n->info = (*proot)->info;
						(*proot)->info = XPF_INFO_NULL;
						n->and = (*proot)->and;
						(*proot)->and = NULL;	/* error if we reference it: for multivalue filter, no-match leads to ->or directly */
						e->and = n;
						n->refcount ++;	/* refcount of (*proot) is unchanged */
						n->or = (*proot)->or;
						if (n->or != NULL)
							n->or->refcount ++;
					}

					goto notidenticalI;	/* save a Get Operation and replication of code! */
				}
			}
			/*
			 * The assertion proot != NULL still holds but e->and can be NULL: must check here.
			 * Now, search hash-table for the place where srcfilter->value can occur. Must
			 * increment (*destfilter's) refcount ALSO: not enough to increment e->refcount!
			 * (see the equivalent delete procedure and the READJUST macro).
			 */
			else {
				pe = xpfGetHashEntry( (HashEntry **)*(unsigned long *)((*destfilter)->value), index, srcfilter->value, size);
				if (*pe == NULL) {
				notidenticalI:
					/* We know that the value is not equal so insert it into a fresh hash-table-entry */
					xpfCreateHashEntry(e, srcfilter->value, size);
					e->nodetype |= (*proot)->nodetype & XPF_DAGS_PROCESS;
					e->and = NULL;
					xpfInsertHashEntry((HashEntry **)*(unsigned long *)((*destfilter)->value), e, index);
					(*destfilter)->refcount ++;	/* due to e */

					if (srcfilter->next == NULL) {	/* fresh guy, no need to sort and insert */
						NOPRINT(xpfPrintf("\t\t\tCOPY dynlist\n");)
						d = srcpattern->programlist;
						pd = &e->common.hash.programlist;
						while (d != NULL) {
							*pd = (XPFProgram *)xpfMalloc(sizeof(XPFProgram));
							xpfMemcpy(&((*pd)->extractor), &(d->extractor), sizeof(XPFExtractor));
							(*pd)->extractor.datatype = xpfAbs(d->extractor.datatype);
							(*pd)->refcount = 1 /* xpfAbs(d->refcount) */;
							pd = &(*pd)->next;
							d = d->next;
						}
						*pd = NULL;

						(*proot)->common.pattern.flags = srcpattern->flags & flagmask;
						e->common.hash.alarm = (*proot)->common.pattern.alarm = srcpattern->alarm;
						e->common.hash.processnum = (*proot)->common.pattern.processnum = (flagmask == XPF_FLAG_USER) ? 0 : srcpattern->processnum;

						if (srcpattern->next == NULL) return e;
						else {
							if ((*proot)->common.pattern.flags & XPF_FLAG_NOSPECIFIC) return NULL;
							else return xpfInsertPatternList(srcpattern->next,
											srcpattern->next->filterlist,
											&e->and,
											NULL,
											NULL,
											granularity, 
											XPF_DAGS_SRCNEWDESTNEW,
											flagmask);
						}
					}
					else {
						srcpattern->header.headertype |= XPF_HEADER_CURRENT;	/* Split in the middle of srcpattern */
						if ((*proot)->common.pattern.flags & XPF_FLAG_NOSPECIFIC) return NULL;
						else return xpfInsertPatternList(srcpattern,
										srcfilter->next,
										&e->and,
										NULL,
										NULL,
										granularity,
										XPF_DAGS_SRCOLDDESTNEW,
										flagmask);
					}
				}
				else {
					(*pe)->refcount ++;
					(*destfilter)->refcount ++;
					if (srcfilter->next == NULL) {
						XPFProgram	*temp;

						/* Don't need "if ((*destfilter)->next != NULL)" since destfilter is a multivalued filter! */

						/*
						 * Here, srcpattern->programlist!=NULL: we have to sort and insert smallest field 1st.
						 * We also have a provision to eliminate identical programlist elements: they are ordered
						 * in the same way as filter elements: greater mask-program elements compared first.
						 * We however make or attempt to identify common portions between program elements (as
						 * the "value" part of a program is undefined). We have to copy "flags" into proot's pattern.
						 */

						NOPRINT(xpfPrintf("\t\t\tCOPY dynlist\n");)
						d = srcpattern->programlist;
						while (d != NULL) {
#if	COUNTUNIT
							if (d->extractor.selector.countunit <= 0) goto endloop1;
#endif	COUNTUNIT
							pd = &(*pe)->common.hash.programlist;
							while (*pd != NULL) {
								if (((*pd)->extractor.selector.level < d->extractor.selector.level) ||
								    (((*pd)->extractor.selector.level == d->extractor.selector.level) &&
									(( (*pd)->extractor.selector.offset XPF_USER_GETCOUNTUNIT((*pd)) ) <
									 ( d->extractor.selector.offset XPF_USER_GETCOUNTUNIT(d)) )))
									pd = &((*pd)->next);
								else if (((*pd)->extractor.selector.level > d->extractor.selector.level) ||
								    (((*pd)->extractor.selector.level == d->extractor.selector.level) &&
									(( (*pd)->extractor.selector.offset XPF_USER_GETCOUNTUNIT((*pd))) >
									 ( d->extractor.selector.offset XPF_USER_GETCOUNTUNIT(d)) )))
									break;
								else { /* they are equal */
									if (*(unsigned long *)(*pd)->extractor.mask == *(unsigned long *)d->extractor.mask) {
										(*pd)->refcount ++ /* += xpfAbs(d->refcount)*/;
										goto endloop1;
									}
									else if (*(unsigned long *)(*pd)->extractor.mask < *(unsigned long *)d->extractor.mask)
										break;
									else pd = &(*pd)->next;
								}
							}
							temp = *pd;
							*pd = (XPFProgram *)xpfMalloc(sizeof(XPFProgram));
							xpfMemcpy(&((*pd)->extractor), &(d->extractor), sizeof(XPFExtractor));
							(*pd)->extractor.datatype = xpfAbs(d->extractor.datatype);
							(*pd)->refcount = 1 /* xpfAbs(d->refcount) */;
							(*pd)->next = temp;	/* I can't assume dynlist is SORTED, so I can't start *pd from CURR. pos */

						endloop1:
							d = d->next;
						}

						(*proot)->common.pattern.flags = srcpattern->flags & flagmask;
						(*pe)->common.hash.alarm = (*proot)->common.pattern.alarm = srcpattern->alarm;
						(*pe)->common.hash.processnum = (*proot)->common.pattern.processnum = (flagmask == XPF_FLAG_USER) ? 0 : srcpattern->processnum;

						if (srcpattern->next != NULL) {
							if ((*proot)->common.pattern.flags & XPF_FLAG_NOSPECIFIC) return NULL;
							else if ((*pe)->and == NULL)
								return xpfInsertPatternList(srcpattern->next,
												srcpattern->next->filterlist,
												&(*pe)->and,
												NULL,
												NULL,
												granularity,
												XPF_DAGS_SRCNEWDESTNEW,
												flagmask);
							else return xpfInsertPatternList(srcpattern->next,
											srcpattern->next->filterlist,
											&(*pe)->and,
											&(*pe)->and->common.pattern.filterlist,
											granularity,
											XPF_DAGS_SRCNEWDESTNEW,
											flagmask);
						}
						else return *pe;
					}
					else {
						srcpattern->header.headertype |= XPF_HEADER_CURRENT;	/* Split in the middle of srcpattern */
						if ((*proot)->common.pattern.flags & XPF_FLAG_NOSPECIFIC) return NULL;
						else if ((*pe)->and == NULL)
							return xpfInsertPatternList(srcpattern,
											srcfilter->next,
											&(*pe)->and,
											NULL,
											NULL,
											granularity,
											XPF_DAGS_SRCOLDDESTNEW,
											flagmask);
						else return xpfInsertPatternList(srcpattern,
										srcfilter->next,
										&(*pe)->and,
										&(*pe)->and->common.pattern.filterlist,
										granularity,
										XPF_DAGS_SRCOLDDESTNEW,
										flagmask);
					}
				}
			}
		}
		else if (ret == XPF_DAGS_SAME) {
		identicalI:
			NOPRINT(xpfPrintf("\t\tXPF_DAGS_SAME advance\n"));
			(*destfilter)->refcount ++ /* += xpfAbs(srcfilter->refcount)*/;

			if (srcfilter->next == NULL) {
				XPFProgram	*temp;

				if ((*destfilter)->next != NULL) {
					/*
					 * The pattern in proot has to be split here with the program hanging off at the already matched
					 * portion and the rest of the portion being in the ->and portion of *proot.
					 */
					xpfCreatePATNode(n);
					n->nodetype |= (*proot)->nodetype & XPF_DAGS_PROCESS;
					n->common.pattern.flags = (*proot)->common.pattern.flags;
					(*proot)->common.pattern.flags &= XPF_FLAG_RESIDUE;	/* NOSPECIFIC and DELAY are taken over by the new-guy */
					n->common.pattern.alarm = (*proot)->common.pattern.alarm;
					n->common.pattern.processnum = (*proot)->common.pattern.processnum;
					n->common.pattern.header.headertype /*|*/= XPF_HEADER_CURRENT;
					n->common.pattern.filterlist = (*destfilter)->next;
					(*destfilter)->next = NULL;
					n->common.pattern.programlist = (*proot)->common.pattern.programlist;
					(*proot)->common.pattern.programlist = NULL;

					n->and = (*proot)->and;
					(*proot)->and = n;
					n->refcount ++;	/* refcount of (*proot)->and is unchanged */
					n->or = (*proot)->or;
				}

				/*
				 * Here, proot's pattern's->programlist!=NULL: we have to sort and insert smallest field 1st.
				 * We also have a provision to eliminate identical programlist elements: they are ordered
				 * in the same way as filter elements: greater mask-program elements compared first.
				 * We however make or attempt to identify common portions between program elements (as
				 * the "value" part of a program is undefined). We have to copy "flags" too.
				 */

				NOPRINT(xpfPrintf("\t\t\tCOPY dynlist\n");)
				d = srcpattern->programlist;
				while (d != NULL) {
#if	COUNTUNIT
					if (d->extractor.selector.countunit <= 0) goto endloop2;
#endif	COUNTUNIT
					pd = &(*proot)->common.pattern.programlist;
					while (*pd != NULL) {
						if (((*pd)->extractor.selector.level < d->extractor.selector.level) ||
						    (((*pd)->extractor.selector.level == d->extractor.selector.level) &&
							(( (*pd)->extractor.selector.offset XPF_USER_GETCOUNTUNIT((*pd))) <
						         ( d->extractor.selector.offset XPF_USER_GETCOUNTUNIT(d)) )))
							pd = &((*pd)->next);
						else if (((*pd)->extractor.selector.level > d->extractor.selector.level) ||
						    (((*pd)->extractor.selector.level == d->extractor.selector.level) &&
							(( (*pd)->extractor.selector.offset XPF_USER_GETCOUNTUNIT((*pd))) >
							 ( d->extractor.selector.offset XPF_USER_GETCOUNTUNIT(d)) )))
							break;
						else { /* they are equal */
							if (*(unsigned long *)(*pd)->extractor.mask == *(unsigned long *)d->extractor.mask) {
								(*pd)->refcount ++ /* += xpfAbs(d->refcount)*/;
								goto endloop2;
							}
							else if (*(unsigned long *)(*pd)->extractor.mask < *(unsigned long *)d->extractor.mask)
								break;
							else pd = &(*pd)->next;
						}
					}
					temp = *pd;
					*pd = (XPFProgram *)xpfMalloc(sizeof(XPFProgram));
					xpfMemcpy(&((*pd)->extractor), &(d->extractor), sizeof(XPFExtractor));
					(*pd)->extractor.datatype = xpfAbs(d->extractor.datatype);
					(*pd)->refcount = 1 /* xpfAbs(d->refcount) */;
					(*pd)->next = temp;	/* I can't assume dyn list is SORTED, so I can't start *pd from CURR. pos */

				endloop2:
					d = d->next;
				}

				(*proot)->common.pattern.flags = srcpattern->flags & flagmask;
				(*proot)->common.pattern.alarm = srcpattern->alarm;
				(*proot)->common.pattern.processnum = (flagmask == XPF_FLAG_USER) ? 0 : srcpattern->processnum;

				if (srcpattern->next == NULL) return *proot;
				else return xpfInsertPatternList(srcpattern->next,
								srcpattern->next->filterlist,
								proot,
								&(*destfilter)->next,
								granularity,
								XPF_DAGS_SRCNEWDESTOLD,
								flagmask);
			}
			else {
				srcpattern->header.headertype |= XPF_HEADER_CURRENT;	/* Split in the middle of srcpattern */
				return xpfInsertPatternList(srcpattern,
							srcfilter->next,
							proot,
							&(*destfilter)->next,
							granularity,
							XPF_DAGS_SRCOLDDESTOLD,
							flagmask);
			}
		}
		else if (ret == XPF_DAGS_ERROR) return NULL;
		else {
			if (commonprefix > 0) {	/* handled differently even if XPF_DAGS_AFTER/BEFORE is known since keys themselves can change */
				NOPRINT(xpfPrintf("\t\tCOMPREFIX: adjust\n");)
				if (destbytesleft > 0) {
					XPFFilter *f = (XPFFilter *)xpfMalloc(sizeof(XPFFilter));

					f->next = (*destfilter)->next;
					(*destfilter)->next = f;

					xpfMemcpy(f->value, (*destfilter)->value + commonprefix, destbytesleft);
					xpfMemcpy(f->extractor.mask, (*destfilter)->extractor.mask + commonprefix, destbytesleft);
					xpfBzero((char *)(f->value + destbytesleft), commonprefix);
					xpfBzero((char *)(f->extractor.mask + destbytesleft), commonprefix);
					f->extractor.datatype = destbytesleft;
					f->extractor.selector.offset =
						(*destfilter)->extractor.selector.offset XPF_USER_GETCOUNTUNIT((*destfilter)) + commonprefix;
#if	COUNTUNIT
					f->extractor.selector.countunit = 1;
#endif	COUNTUNIT
					f->extractor.selector.level = 0;
					f->refcount = 1 /* (*destfilter)->refcount */;

					xpfBzero((char *)((*destfilter)->value + commonprefix), destbytesleft);
					xpfBzero((char *)((*destfilter)->extractor.mask + commonprefix), destbytesleft);
					(*destfilter)->extractor.datatype = commonprefix;
					(*destfilter)->refcount ++ /* += xpfAbs(srcfilter->refcount) */;
				}
				destfilter = &(*destfilter)->next;

				if (srcbytesleft > 0) {
					xpfMemcpy(srcfilter->value, srcfilter->value + commonprefix, srcbytesleft);
					xpfMemcpy(srcfilter->extractor.mask, srcfilter->extractor.mask + commonprefix, srcbytesleft);
					xpfBzero((char *)(srcfilter->value + srcbytesleft), commonprefix);
					xpfBzero((char *)(srcfilter->extractor.mask + srcbytesleft), commonprefix);
					srcfilter->extractor.datatype = srcbytesleft;
					srcfilter->extractor.selector.offset =
						srcfilter->extractor.selector.offset XPF_USER_GETCOUNTUNIT(srcfilter) + commonprefix;
#if	COUNTUNIT
					srcfilter->extractor.selector.countunit = 1;
#endif	COUNTUNIT
				}
				else srcfilter = srcfilter->next;

				if (srcfilter == NULL) {
					if (srcpattern->next == NULL) return *proot;
					else return xpfInsertPatternList(srcpattern->next,
									srcpattern->next->filterlist,
									proot,
									destfilter,
									granularity,
									XPF_DAGS_SRCNEWDESTOLD,
									flagmask);
				}
				else return xpfInsertPatternList(srcpattern,
								srcfilter,
								proot,
								destfilter,
								granularity,
								XPF_DAGS_SRCOLDDESTOLD,
								flagmask);
			}
			else if (ret == XPF_DAGS_AFTER) {
				/*
				 * Very similar to CASE 2: I have to insert source after the destination.
				 */
			case2I:
				NOPRINT(xpfPrintf("\t\tXPF_DAGS_AFTER: advance\n");)
				if ((*proot)->common.pattern.flags & XPF_FLAG_NOSPECIFIC) return NULL;	/* processes aren't effected since their flags are 0 */

				srcpattern->header.headertype |= XPF_HEADER_CURRENT;	/* src+dest were in same hdr => src's header shouldn't be extracted */
				if (*destfilter != (*proot)->common.pattern.filterlist) {	/* create new node to hold matched portion so far */
					xpfCreatePATNode(n);
					n->nodetype |= (*proot)->nodetype & XPF_DAGS_PROCESS;
					n->common.pattern.flags = (*proot)->common.pattern.flags;
					(*proot)->common.pattern.flags &= XPF_FLAG_RESIDUE;	/* NOSPECIFIC and DELAY are taken over by the new-guy */
					n->common.pattern.alarm = (*proot)->common.pattern.alarm;
					n->common.pattern.processnum = (*proot)->common.pattern.processnum;
					n->common.pattern.header.headertype /*|*/= XPF_HEADER_CURRENT;
					n->common.pattern.filterlist = *destfilter;
					*destfilter = NULL;
					n->common.pattern.programlist = (*proot)->common.pattern.programlist;
					(*proot)->common.pattern.programlist = NULL;

					n->and = (*proot)->and;
					(*proot)->and = n;
					n->refcount ++;	/* refcount of (*proot)->and is unchanged */
					n->or = (*proot)->or;
					if (n->or != NULL) {
						n->or->refcount ++;
						return xpfInsertPatternList(srcpattern,
										srcfilter,
										&n->or,
										&(n->or->common.pattern.filterlist),
										granularity,
										XPF_DAGS_SRCOLDDESTNEW,
										flagmask);
					}
					else return xpfInsertPatternList(srcpattern,
									srcfilter,
									&n->or,
									NULL,
									NULL,
									granularity,
									XPF_DAGS_SRCOLDDESTNEW,
									flagmask);
				}
				else if ((*proot)->or != NULL) return xpfInsertPatternList(srcpattern,
											srcfilter,
											&(*proot)->or,
											&((*proot)->or->common.pattern.filterlist),
											granularity,
											XPF_DAGS_SRCOLDDESTNEW,
											flagmask);
				else return xpfInsertPatternList(srcpattern,
								srcfilter,
								&(*proot)->or,
								NULL,
								NULL,
								granularity,
								XPF_DAGS_SRCOLDDESTNEW,
								flagmask);

			}
			else if (ret == XPF_DAGS_BEFORE) {
				/*
				 * Here, n is inserted before m in decision tree at proot. Source has to go before destination => end of recursion.
				 */
			case3I:
				NOPRINT(xpfPrintf("\t\tXPF_DAGS_BEFORE: terminate\n");)
				if (*destfilter != (*proot)->common.pattern.filterlist) {
					srcpattern->header.headertype |= XPF_HEADER_CURRENT;	/* srcpattern is AFTER a part of *proot's pattern */
				}
				else {
					(*proot)->common.pattern.header.headertype |= XPF_HEADER_CURRENT;	/* both have = headers, so extract only ONCE */
				}

				xpfCreatePATNode(n);	/* create new node to hold the current srcpattern */
				n->nodetype |= srcpattern->flags & flagmask & XPF_DAGS_PROCESS;
				n->common.pattern.flags = srcpattern->flags & flagmask;
				n->common.pattern.alarm = srcpattern->alarm;
				n->common.pattern.processnum = (flagmask == XPF_FLAG_USER) ? 0 : srcpattern->processnum;
#if	VARHEADER
				n->common.pattern.header.size.variable = NULL;
#endif	VARHEADER
				COPY_HEADER(&(n->common.pattern.header), &(srcpattern->header));

				n->common.pattern.filterlist = (XPFFilter *)xpfMalloc(sizeof(XPFFilter));
				xpfMemcpy(&(n->common.pattern.filterlist->extractor), &(srcfilter->extractor), sizeof(XPFExtractor));
				n->common.pattern.filterlist->extractor.datatype = xpfAbs(srcfilter->extractor.datatype);
				xpfMemcpy(n->common.pattern.filterlist->value, srcfilter->value, XPF_DATA_MAXSIZE);
				n->common.pattern.filterlist->refcount = 1 /* xpfAbs(srcfilter->refcount) */;
				n->common.pattern.filterlist->next = NULL;
				n->common.pattern.programlist = NULL;
				n->common.pattern.next = NULL;

				if (*destfilter != (*proot)->common.pattern.filterlist) {	/* create new node to hold un-matched portion of proot's pat */
					xpfCreatePATNode(m);
					m->nodetype |= (*proot)->nodetype & XPF_DAGS_PROCESS;
					m->common.pattern.flags = (*proot)->common.pattern.flags;
					(*proot)->common.pattern.flags &= XPF_FLAG_RESIDUE;	/* NOSPECIFIC and DELAY are taken over by the new-guy */
					m->common.pattern.alarm = (*proot)->common.pattern.alarm;
					m->common.pattern.processnum = (*proot)->common.pattern.processnum;
					COPY_HEADER(&m->common.pattern.header, &(*proot)->common.pattern.header);
					m->common.pattern.header.headertype |= XPF_HEADER_CURRENT;	/* always CURRENT */
					m->common.pattern.filterlist = *destfilter;
					*destfilter = NULL;
					m->common.pattern.programlist = (*proot)->common.pattern.programlist;
					(*proot)->common.pattern.programlist = NULL;

					m->and = (*proot)->and;
					(*proot)->and = n;
					retval = &(*proot)->and;
					n->refcount ++;
					m->or = (*proot)->or;
					if ((*proot)->or != NULL) (*proot)->or->refcount ++;
					n->and = NULL;
					n->or = m;
					m->refcount ++;
				}
				else {
					n->or = *proot;
					*proot = n;
					retval = proot;
					n->refcount ++;
					n->and = NULL;
				}
				/* n->and is NULL in both cases */

				if (srcfilter->next == NULL) {
					if (srcpattern->next == NULL) {	/* Have to copy the programlist */
						NOPRINT(xpfPrintf("\t\t\tCOPY dynlist\n");)
						if (srcpattern->programlist != NULL) {
							DOPRINT(xpfAssert((*retval)->common.pattern.programlist == NULL));
							pd = &(*retval)->common.pattern.programlist;
							d = srcpattern->programlist;
							while (d != NULL) {
								*pd = (XPFProgram *)xpfMalloc(sizeof(XPFProgram));
								xpfMemcpy(&((*pd)->extractor), &(d->extractor), sizeof(XPFExtractor));
								(*pd)->extractor.datatype = xpfAbs(d->extractor.datatype);
								(*pd)->refcount = 1 /* xpfAbs(d->refcount) */;
								pd = &(*pd)->next;
								d = d->next;
							}
							*pd = NULL;
						}
						return *retval;
					}
					else return xpfInsertPatternList(srcpattern->next,
									srcpattern->next->filterlist,
									&n->and,
									NULL,
									(n->nodetype & XPF_DAGS_PROCESS ? (n->or) : NULL),	/* shady way to pass OR info */
									granularity,
									XPF_DAGS_SRCNEWDESTNEW,
									flagmask);
				}
				else {
					/* There is absolutely no need to go to n->and since it is NULL anyway: reduce #of nodes, also bug in Launch */
					srcpattern->header.headertype |= XPF_HEADER_CURRENT;	/* Split in the middle of srcpattern */
					return xpfInsertPatternList(srcpattern,
								srcfilter->next,
								&n->and,
								NULL,
								(n->nodetype & XPF_DAGS_PROCESS ? (n->or) : NULL),	/* shady way to pass OR info */
								granularity,
								XPF_DAGS_SRCOLDDESTNEW,
								flagmask);
				}
			}
			else return NULL;	/* neither SAME not BEFORE nor AFTER: must be ERROR */
		}
	}
	else if ( (location == XPF_DAGS_SRCNEWDESTOLD) ||
		  ((location == XPF_DAGS_SRCNEWDESTNEW) && ((*proot)->common.pattern.header.headertype & XPF_HEADER_CURRENT)) )
	{
		/*
		 * CASE 2: srcfilter is in a new header, destfilter is still in the old header.
		 * Current protocol always has a higher priority over another protocol.
		 * Control comes here also when srcpattern (as a whole) is AFTER proot's pattern.
		 */
		NOPRINT(xpfPrintf("\tSRC=new, DEST=old\n");)
		if ((*proot)->common.pattern.flags & XPF_FLAG_NOSPECIFIC) return NULL;	/* processes aren't effected since their flags are 0 */

		if (*destfilter != (*proot)->common.pattern.filterlist) {
			NOPRINT(xpfPrintf("\t\tCreate new dest-node\n");)
			xpfCreatePATNode(n);
			n->nodetype |= (*proot)->nodetype & XPF_DAGS_PROCESS;
			n->common.pattern.flags = (*proot)->common.pattern.flags;
			(*proot)->common.pattern.flags &= XPF_FLAG_RESIDUE;	/* NOSPECIFC and DELAY are taken over by the new guy */
			n->common.pattern.alarm = (*proot)->common.pattern.alarm;
			n->common.pattern.processnum = (*proot)->common.pattern.processnum;
			n->common.pattern.header.headertype /*|*/= XPF_HEADER_CURRENT;
			n->common.pattern.filterlist = *destfilter;
			*destfilter = NULL;
			n->common.pattern.programlist = (*proot)->common.pattern.programlist;
			(*proot)->common.pattern.programlist = NULL;

			n->and = (*proot)->and;
			(*proot)->and = n;
			n->refcount ++;	/* refcount of (*proot)->and remains unchanged */
			n->or = (*proot)->or;
			if (n->or != NULL) {
				n->or->refcount ++;
				return xpfInsertPatternList(srcpattern,
							srcfilter,
							&n->or,
							&(n->or->common.pattern.filterlist),
							granularity,
							XPF_DAGS_SRCOLDDESTNEW,
							flagmask);
			}
			else return xpfInsertPatternList(srcpattern,
							srcfilter,
							&n->or,
							NULL,
							NULL,
							granularity,
							XPF_DAGS_SRCOLDDESTNEW,
							flagmask);
		}
		else {
			if ((*proot)->or == NULL) return xpfInsertPatternList(srcpattern,
										srcfilter,
										&(*proot)->or,
										NULL,
										NULL,
										granularity,
										XPF_DAGS_SRCOLDDESTNEW,
										flagmask);
			else return xpfInsertPatternList(srcpattern,
							srcfilter,
							&(*proot)->or,
							&((*proot)->or->common.pattern.filterlist),
							granularity,
							XPF_DAGS_SRCOLDDESTNEW,
							flagmask);
		}
	}
	else if ( (location == XPF_DAGS_SRCOLDDESTNEW) ||
		  ((location == XPF_DAGS_SRCNEWDESTNEW) && (srcpattern->header.headertype & XPF_HEADER_CURRENT)) )
	{
		/*
		 * CASE 3: srcfilter is not in a new header, but destfilter is.
		 * Control comes here also when srcpattern (as a whole) is BEFORE proot's pattern.
		 */
		NOPRINT(xpfPrintf("\tSRC=old, DEST=new\n");)
		if ((location == XPF_DAGS_SRCNEWDESTNEW) && (srcpattern->flags & XPF_FLAG_PROCESS) && ((*proot)->common.pattern.flags & XPF_FLAG_PROCESS)) {
			/* Both are processes and src (new process with current header of generating pattern) has to be inserted at the same node as dest */
			srcpattern->header.headertype &= ~XPF_HEADER_CURRENT;
			goto case4I;
		}

		xpfCreatePATNode(n);
		n->nodetype |= srcpattern->flags & flagmask & XPF_DAGS_PROCESS;
		n->common.pattern.flags = srcpattern->flags & flagmask;
		n->common.pattern.alarm = srcpattern->alarm;
		n->common.pattern.processnum = (flagmask == XPF_FLAG_USER) ? 0 : srcpattern->processnum;
#if	VARHEADER
		n->common.pattern.header.size.variable = NULL;
#endif	VARHEADER
		COPY_HEADER(&(n->common.pattern.header), &(srcpattern->header));
		srcpattern->header.headertype |= XPF_HEADER_CURRENT;	/* always CURRENT since it always goes BEFORE *proot's pattern */

		ps = &(n->common.pattern.filterlist);
		while (srcfilter != NULL) {
			*ps = (XPFFilter *)xpfMalloc(sizeof(XPFFilter));
			xpfMemcpy(&((*ps)->extractor), &(srcfilter->extractor), sizeof(XPFExtractor));
			(*ps)->extractor.datatype = xpfAbs(srcfilter->extractor.datatype);
			xpfMemcpy((*ps)->value, srcfilter->value, XPF_DATA_MAXSIZE);
			(*ps)->refcount = 1 /* xpfAbs(srcfilter->refcount) */;
			ps = &(*ps)->next;
			srcfilter = srcfilter->next;
		}
		*ps = NULL;
		n->common.pattern.programlist = NULL;

		n->or = (*proot);
		*proot = n;
		n->refcount ++;
		n->and = NULL;

		DOPRINT(xpfAssert(srcfilter == NULL));
		if (srcpattern->next == NULL) {

			NOPRINT(xpfPrintf("\t\tCOPY dynlist\n");)
			pd = &n->common.pattern.programlist;
			DOPRINT(xpfAssert(*pd == NULL));

			d = srcpattern->programlist;
			while (d != NULL) {
				*pd = (XPFProgram *)xpfMalloc(sizeof(XPFProgram));
				xpfMemcpy(&((*pd)->extractor), &(d->extractor), sizeof(XPFExtractor));
				(*pd)->extractor.datatype = xpfAbs(d->extractor.datatype);
				(*pd)->refcount = 1 /*xpfAbs(d->refcount) */;
				pd = &(*pd)->next;
				d = d->next;
			}
			*pd = NULL;

			return *proot;
		}
		else return xpfInsertPatternList(srcpattern->next,
						srcpattern->next->filterlist,
						&n->and,
						NULL,
						NULL,
						granularity,
						XPF_DAGS_SRCNEWDESTNEW,
						flagmask);
	}
	else {
		/*
		 * CASE 4: Both are in new headers. The headers must match exactly.
		 */
	case4I:
		NOPRINT(xpfPrintf("\tNEW headers\n");)
		if (srcpattern->header.headertype != (*proot)->common.pattern.header.headertype)
			return NULL;
#if	VARHEADER
		if (srcpattern->header.headertype & XPF_HEADER_VARIABLE) {
			if (bcmp(srcpattern->header.size.variable, (*proot)->common.pattern.header.size.variable, sizeof(XPFVariableHeader)) != 0) {
				ERRPRINT(xpfPrintf("Attempt to change header of pattern!\n"));
				return NULL;
			}
		}
		else
#endif	VARHEADER
		if (srcpattern->header.size.fixed != (*proot)->common.pattern.header.size.fixed)
			return NULL;
		goto case1I;
	}
}

/*
 * Inserts node into the subtree starting at proot. Also does some cleanup
 * whenever necessary, which is similar to the READJUST-macro. Updates the
 * deletedmsg/info * list only when there is an error. Otherwise, it merges
 * the deleted messages/information with those at existing PATNode-s.
 * This is called for internal adjustments only: never in direct response
 * to a user-specified operation.
 */
int
xpfInsertPATNode(proot, node, granularity
#if	MERGEINFO
		, deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
		, deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
		)
	PATNode		**proot;
	PATNode		*node;
	int		granularity;
#if	MERGEINFO
        XPFInfoList     **deletedinfo;
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
        XPFMessageList  **deletedmessagelist;
#endif	DELAYEDMSGS && STOREMSGS
{
	PATNode		*temp, *retval;

	if (node == NULL) return TRUE;
	if (*proot == NULL) {
		*proot = node;
		node->refcount ++;
		return TRUE;
	}
	if ((*proot) == node) return TRUE;	/* can happen since dag and refcounts can be > 1 */
	if ((*proot)->common.pattern.filterlist == NULL) {
		xpfDestroyProgramList((*proot)->common.pattern.programlist);
		temp = *proot;
		*proot = node;
		node->refcount += temp->refcount;	/* all those guys now point to node */
		temp->refcount = 0;
		xpfDestroyDeadPATTree(&temp
#if	MERGEINFO
					, &node->info
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
					, &node->messagelist
#endif	DELAYEDMSGS && STOREMSGS
					);
		return TRUE;
	}

	if ((retval = xpfInsertPatternList(node->common.pattern,
					node->common.pattern.filterlist,
					proot,
					(*proot) == NULL ? NULL : &(*proot)->common.pattern.filterlist,
					granularity,
					XPF_DAGS_SRCNEWDESTNEW,
					XPF_FLAG_DAGS)) == NULL) {
		/* Do I undo the effects of insertion by a delete ? Maybe not since a part of it could be inserted but the whole can be deleted! */
		node->refcount = 0;
		xpfDestroyDeadPATTree(&node
#if	MERGEINFO
					, deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
					, deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
					);
		return FALSE;
	}
	else {
		if (!xpfInsertPATNode(&retval->or, node->or, granularity
#if	MERGEINFO
					, deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
					, deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
					)) {
			xpfDestroyPATNode(&node
#if	MERGEINFO
					, deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
					, deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
					);
			return FALSE;
		}
		if (!xpfInsertPATNode(&retval->and, node->and, granularity
#if	MERGEINFO
					, deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
					, deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
					)) {
			xpfDestroyPATNode(&node
#if	MERGEINFO
					, deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
					, deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
					);
			return FALSE;
		}
		xpfDestroyPATNode(&node
#if	MERGEINFO
				, deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
				, deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
				);
		return TRUE;
	}
}

/* The part of the ~FILTER macro (below) that can be used in both DeletePatternList and TerminateProcesses */
#if	DELAYEDMSGS && STOREMSGS
#if	MERGEINFO
#define READJUST_AFTER_DELETE_PATTERN() \
{\
	PATNode *temp1, *temp2;\
\
	if ( ((*proot)->and == NULL) && ((*proot)->or == NULL) ) {\
		xpfDestroyPATNode(proot, deletedinfo, deletedmessagelist);\
	}\
	else if ((*proot)->or == NULL) {\
		temp2 = *proot;\
		xpfDestroyPATNode(&temp2, deletedinfo, deletedmessagelist);\
		*proot = (*proot)->and;\
	}\
	else if ((*proot)->and == NULL) {\
		temp2 = *proot;\
		xpfDestroyPATNode(&temp2, deletedinfo, deletedmessagelist);\
		*proot = (*proot)->or;\
	}\
	else if (!((*proot)->and->common.pattern.header.headertype & XPF_HEADER_CURRENT) && \
		 !((*proot)->or->common.pattern.header.headertype & XPF_HEADER_CURRENT)) {\
		/* IP gone w/o TCP/UDP going: this is serious trouble */ \
		xpfDestroyDeadPATTree(proot, deletedinfo, deletedmessagelist);\
		*proot = NULL;\
	}\
	/* else: non-tirivial case: you have to delete (*proot) and merge its and/or sons in some way: proot still has or useful info */\
	else if ((*proot)->and->common.pattern.header.headertype & XPF_HEADER_CURRENT) {\
		temp1 = (*proot)->or;\
		temp2 = *proot;\
		xpfDestroyPATNode(&temp2, deletedinfo, deletedmessagelist);\
		*proot = (*proot)->and;\
		xpfInsertPATNode(proot, temp1, granularity);\
	}\
	else { /* the or portion has to be in the same header as proot: can this ever occur since old hdr is always @ a higher priority than new */\
		temp1 = (*proot)->and;\
		temp2 = *proot;\
		xpfDestroyPATNode(&temp2, deletedinfo, deletedmessagelist);\
		*proot = (*proot)->or;\
		xpfInsertPATNode(proot, temp1, granularity);\
	}\
}
#else	MERGEINFO
#define READJUST_AFTER_DELETE_PATTERN() \
{\
	PATNode *temp1, *temp2;\
\
	if ( ((*proot)->and == NULL) && ((*proot)->or == NULL) ) {\
		xpfDestroyPATNode(proot, deletedmessagelist);\
	}\
	else if ((*proot)->or == NULL) {\
		temp2 = *proot;\
		xpfDestroyPATNode(&temp2, deletedmessagelist);\
		*proot = (*proot)->and;\
	}\
	else if ((*proot)->and == NULL) {\
		temp2 = *proot;\
		xpfDestroyPATNode(&temp2, deletedmessagelist);\
		*proot = (*proot)->or;\
	}\
	else if (!((*proot)->and->common.pattern.header.headertype & XPF_HEADER_CURRENT) && \
		 !((*proot)->or->common.pattern.header.headertype & XPF_HEADER_CURRENT)) {\
		/* IP gone w/o TCP/UDP going: this is serious trouble */ \
		xpfDestroyDeadPATTree(proot, deletedmessagelist);\
		*proot = NULL;\
	}\
	/* else: non-tirivial case: you have to delete (*proot) and merge its and/or sons in some way */\
	else if ((*proot)->and->common.pattern.header.headertype & XPF_HEADER_CURRENT) {\
		temp1 = (*proot)->or;\
		temp2 = *proot;\
		xpfDestroyPATNode(&temp2, deletedmessagelist);\
		*proot = (*proot)->and;\
		xpfInsertPATNode(proot, temp1, granularity);\
	}\
	else { /* the or portion has to be in the same header as proot: can this ever occur since old hdr is always @ a higher priority than new */\
		temp1 = (*proot)->and;\
		temp2 = *proot;\
		xpfDestroyPATNode(&temp2, deletedmessagelist);\
		*proot = (*proot)->or;\
		xpfInsertPATNode(proot, temp1, granularity);\
	}\
}
#endif	MERGEINFO
#else	DELAYEDMSGS && STOREMSGS
#if	MERGEINFO
#define READJUST_AFTER_DELETE_PATTERN() \
{\
	PATNode *temp1, *temp2;\
\
	if ( ((*proot)->and == NULL) && ((*proot)->or == NULL) ) {\
		xpfDestroyPATNode(proot, deletedinfo);\
	}\
	else if ((*proot)->or == NULL) {\
		temp2 = *proot;\
		xpfDestroyPATNode(&temp2, deletedinfo);\
		*proot = (*proot)->and;\
	}\
	else if ((*proot)->and == NULL) {\
		temp2 = *proot;\
		xpfDestroyPATNode(&temp2, deletedinfo);\
		*proot = (*proot)->or;\
	}\
	else if (!((*proot)->and->common.pattern.header.headertype & XPF_HEADER_CURRENT) && \
		 !((*proot)->or->common.pattern.header.headertype & XPF_HEADER_CURRENT)) {\
		/* IP gone w/o TCP/UDP going: this is serious trouble */ \
		xpfDestroyDeadPATTree(proot, deletedinfo);\
		*proot = NULL;\
	}\
	/* else: non-tirivial case: you have to delete (*proot) and merge its and/or sons in some way: proot still has or useful info */\
	else if ((*proot)->and->common.pattern.header.headertype & XPF_HEADER_CURRENT) {\
		temp1 = (*proot)->or;\
		temp2 = *proot;\
		xpfDestroyPATNode(&temp2, deletedinfo);\
		*proot = (*proot)->and;\
		xpfInsertPATNode(proot, temp1, granularity);\
	}\
	else { /* the or portion has to be in the same header as proot: can this ever occur since old hdr is always @ a higher priority than new */\
		temp1 = (*proot)->and;\
		temp2 = *proot;\
		xpfDestroyPATNode(&temp2, deletedinfo);\
		*proot = (*proot)->or;\
		xpfInsertPATNode(proot, temp1, granularity);\
	}\
}
#else	MERGEINFO
#define READJUST_AFTER_DELETE_PATTERN() \
{\
	PATNode *temp1, *temp2;\
\
	if ( ((*proot)->and == NULL) && ((*proot)->or == NULL) ) {\
		xpfDestroyPATNode(proot);\
	}\
	else if ((*proot)->or == NULL) {\
		temp2 = *proot;\
		xpfDestroyPATNode(&temp2);\
		*proot = (*proot)->and;\
	}\
	else if ((*proot)->and == NULL) {\
		temp2 = *proot;\
		xpfDestroyPATNode(&temp2);\
		*proot = (*proot)->or;\
	}\
	else if (!((*proot)->and->common.pattern.header.headertype & XPF_HEADER_CURRENT) && \
		 !((*proot)->or->common.pattern.header.headertype & XPF_HEADER_CURRENT)) {\
		/* IP gone w/o TCP/UDP going: this is serious trouble */ \
		xpfDestroyDeadPATTree(proot);\
		*proot = NULL;\
	}\
	/* else: non-tirivial case: you have to delete (*proot) and merge its and/or sons in some way */\
	else if ((*proot)->and->common.pattern.header.headertype & XPF_HEADER_CURRENT) {\
		temp1 = (*proot)->or;\
		temp2 = *proot;\
		xpfDestroyPATNode(&temp2);\
		*proot = (*proot)->and;\
		xpfInsertPATNode(proot, temp1, granularity);\
	}\
	else { /* the or portion has to be in the same header as proot: can this ever occur since old hdr is always @ a higher priority than new */\
		temp1 = (*proot)->and;\
		temp2 = *proot;\
		xpfDestroyPATNode(&temp2);\
		*proot = (*proot)->or;\
		xpfInsertPATNode(proot, temp1, granularity);\
	}\
}
#endif	MERGEINFO
#endif	DELAYEDMSGS && STOREMSGS

/*
 * This will be perfectly symmetric if destfilter->refcount was checked INSIDE
 * this function and was freed in a similar way. But we can't do that since this
 * function is used in such a way that ^ operation and program list operation have
 * to be done together BEFORE the DELETE_PATTERN operation. Note that this macro
 * prevents the case when there is a proot with nothing in it so that Insert works
 * properly. Also note that if THIS macro "changes" anything, functions called later
 * DO NOT dereference *proot or *destfilter (so no problem if they become NULL).
 */
#define READJUST_AFTER_DELETE_FILTER()\
{\
	if ((*proot)->common.pattern.filterlist == NULL) { \
		*destfilter = NULL;\
		ret = TRUE;\
		if (--(*proot)->refcount <= 0) {\
			READJUST_AFTER_DELETE_PATTERN();\
		}\
	}\
	if (*proot == NULL) {*destfilter = NULL; }\
	else if (*destfilter == NULL) { destfilter = &(*proot)->common.pattern.filterlist; }\
}

/*
 * Deletes the PAT node from the given DAG and frees memory allocated during insert.
 * Deletion does not interpret "*" (XPF_DATA_NONE), i.e., it looks at filters literally.
 * Changes deletedinfo by assigning it to the deleted info ONLY IF
 * deletion proceeded without any error. Otherwise it remains as it was before.
 * Returns TRUE if successful deletion, FASLE otherwise: cannot return "bad" info
 * since the XPF cannot distinguish between a valid info and an invalid one.
 * It sets the deletedmessagelist in a similar fashion: it is unassigned if error.
 *
 * Assumes that if a node is not NULL, it has atleast one filter in its filterlist.
 */
int
xpfDeletePatternList(srcpattern, srcfilter, proot, destfilter, granularity, location
#if	MERGEINFO
		, deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
		, deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
		)
	XPFPattern	*srcpattern;
	XPFFilter	*srcfilter;	
	PATNode		**proot;
	XPFFilter	**destfilter;
	int		granularity;
	int		location;
#if	MERGEINFO
	XPFInfoList	**deletedinfo;
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
	XPFMessageList	**deletedmessagelist;
#endif	DELAYEDMSGS && STOREMSGS
{
	XPFProgram	*d, **pd;
	int		ret, commonprefix, srcbytesleft, destbytesleft;
	HashEntry	**pe, *te;
	int		size, index;	/* for hash-table search */

	/*
	 * You should check for the possibility that srcfilter / srcpattern can
	 * become NULL before making ANY recursive call.
	 */
	if (srcfilter == NULL || srcpattern == NULL) return FALSE;	/* It was TRUE: I MUST check before every recursive call */

	if (*proot == NULL) return FALSE;	/* could not delete: specified source patternlist non-existent */
	if (*destfilter == NULL) {
		NOPRINT(xpfPrintf("\tGoto NEW proot\n");)
		if ((*proot)->and == NULL) return FALSE;
		else return xpfDeletePatternList(srcpattern,
						srcfilter,
						&((*proot)->and),
						&((*proot)->and->common.pattern.filterlist),
						granularity,
						location | XPF_DAGS_DESTISNEW
#if	MERGEINFO
						,deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
						,deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
						);
	}

	ERRPRINT(xpfAssert((*destfilter != NULL) && (*proot != NULL)));

	if ( (location == XPF_DAGS_SRCOLDDESTOLD) ||
	     ((location == XPF_DAGS_SRCNEWDESTOLD) && (srcpattern->header.headertype & XPF_HEADER_CURRENT)) ||
	     ((location == XPF_DAGS_SRCOLDDESTNEW) && ((*proot)->common.pattern.header.headertype & XPF_HEADER_CURRENT)) ||
	     ((location == XPF_DAGS_SRCNEWDESTNEW) && (srcpattern->header.headertype & XPF_HEADER_CURRENT) && ((*proot)->common.pattern.header.headertype & XPF_HEADER_CURRENT)) )
	{
		/*
		 * CASE 1: Both within the same header as before.
		 */
	case1D:
		if (location == XPF_DAGS_SRCNEWDESTNEW) {
			if ((ret = xpfOrderPatterns(srcpattern, &(*proot)->common.pattern)) == XPF_DAGS_AFTER) goto case2D;
			else if (ret == XPF_DAGS_BEFORE) goto case3D;
			/* else carry on */
		}
		NOPRINT(xpfPrintf("\tOLD headers\n");)
		xpfOrderFilters((srcfilter), (*destfilter), (commonprefix), (srcbytesleft), (destbytesleft), (granularity), (ret), end_order6);
		if(ret == XPF_DAGS_SAMEMASK) {
			DOPRINT(xpfPrintf("\t\tXPF_DAGS_SAMEMASK see hash\n");)

			size = xpfSizeFromType(srcfilter->extractor.datatype);
			index = xpfHashIndex(srcfilter->value);

			if ((*destfilter)->extractor.datatype > 0) {
				if (*(unsigned long *)srcfilter->value == *(unsigned long *)(*destfilter)->value) {
					/* same as XPF_DAGS_SAME with common prefix = size */
					commonprefix = size;
					srcbytesleft = 0;
					destbytesleft = 0;
					goto identicalD;
				}
				else if (*(unsigned long *)srcfilter->value > *(unsigned long *)(*destfilter)->value) {
					commonprefix = 0;
					ret = XPF_DAGS_BEFORE;
					goto notidenticalD;
				}
				else {
					commonprefix = 0;
					ret = XPF_DAGS_AFTER;
					goto notidenticalD;
				}
				/* not reached */
			}
			else {
				pe = xpfGetHashEntry( (HashEntry **)*(unsigned long *)((*destfilter)->value), index, srcfilter->value, size);
				if (*pe == NULL) {
					commonprefix = 0;
					ret = XPF_DAGS_AFTER;
					goto notidenticalD;
					/* cannot be BEFORE else wouldn't have come here: if at all it can occur if same *destfilter REPEATS later */
				}
				else {	/* Only place to DO something: the assertion proot != NULL still holds but *pe->and can be NULL: must check here */
					NOPRINT(xpfPrintf("\t\tXPF_DAGS_SAMEMASK first delete hashentry\n"));
					(*pe)->refcount --;
					(*destfilter)->refcount -- /* -= xpfAbs(srcfilter->refcount) */;
					if ((*pe)->refcount <= 0) {
#if	DELAYEDMSGS && STOREMSGS
						XPFMessageList	**psm;
#endif	DELAYEDMSGS && STOREMSGS
#if	MERGEINFO
						XPFInfoList	**psl;
#endif	MERGEINFO

						xpfDeleteHashEntry(pe);
						if ((*pe)->and != NULL) {
							if ((*destfilter)->refcount <= 0) {	/* aha! this was the 1st hash entry in the table */
								(*proot)->and = (*pe)->and;	/* now can readjust as usual as if nothing has happened */
								xpfDestroyHashEntry(*pe);
								xpfDestroyHashTable((*destfilter));
								ret = FALSE;
								goto restisidenticalD;
							}
							else {
								if (--(*pe)->and->refcount <= 0) {	/* this node cannot have any references outside "pe" */
									xpfDestroyDeadPATTree(&(*pe)->and
#if	MERGEINFO
												, deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
												, deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
												);
								}
							}
						}

#if	MERGEINFO
						/* These two must be saved since the proot with the old association is no longer going to exist */
						psl = &(*pe)->info;
						while (*psl != NULL) psl = &(*psl)->next;
						*psl = *deletedinfo;
						*deletedinfo = (*pe)->info;
						(*pe)->info = XPF_INFO_NULL;
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
						psm = &(*pe)->messagelist;
						while (*psm != NULL) psm = &(*psm)->next;
						*psm = *deletedmessagelist;
						*deletedmessagelist = (*pe)->messagelist;
						(*pe)->messagelist = NULL;
#endif	DELAYEDMSGS && STOREMSGS
						xpfDestroyHashEntry(*pe);
						return TRUE;
					}
					if ((*destfilter)->refcount <= 0) {	/* some error occurred and there are no hash entries */
						xpfDestroyHashTable((*destfilter));
						ret = FALSE;
						goto restisidenticalD;
					}
					/* else, delete rest of srcpattern from pe->and after deleting the programlist from pe if (*destfilter)->next = NULL */
					if (srcfilter->next == NULL) {
						if (srcpattern->next == NULL) {

							NOPRINT(xpfPrintf("\t\tDELETE dynlist\n");)
							 d = srcpattern->programlist;
							 while (d!= NULL) {
#if	COUNTUNIT
								if (d->extractor.selector.countunit <= 0) goto endloop3;
#endif	COUNTUNIT
								pd = &(*pe)->common.hash.programlist;
								while (*pd != NULL) {
									if (( (*pd)->extractor.selector.level != d->extractor.selector.level) ||
									(( (*pd)->extractor.selector.offset XPF_USER_GETCOUNTUNIT((*pd))) !=
									   ( d->extractor.selector.offset XPF_USER_GETCOUNTUNIT(d)) ))
										pd = &(*pd)->next;
									else {
										if (*(unsigned long *)(*pd)->extractor.mask != *(unsigned long *)d->extractor.mask)
											pd = &(*pd)->next;
										else {
											(*pd)->refcount -- /* -= xpfAbs(d->refcount) */;
											if ((*pd)->refcount <= 0) {
												XPFProgram *temp = *pd; *pd = (*pd)->next; xpfFree((char *)temp, sizeof(XPFProgram));
											}
											goto endloop3;
										}
									}
								}

							endloop3:
								d = d->next;
							}
							return TRUE;
						}
						else return xpfDeletePatternList(srcpattern->next,
										srcpattern->next->filterlist,
										&(*pe)->and,
										&(*pe)->and->common.pattern.filterlist,
										granularity,
										XPF_DAGS_SRCNEWDESTNEW
#if	MERGEINFO
										,deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
										,deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
										);
					}
					else {
						srcpattern->header.headertype &= ~XPF_HEADER_CURRENT;	/* proot's pattern is no longer current */
						return xpfDeletePatternList(srcpattern,
									srcfilter->next,
									&(*pe)->and,
									&(*pe)->and->common.pattern.filterlist,
									granularity,
									XPF_DAGS_SRCNEWDESTNEW
#if	MERGEINFO
									,deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
									,deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
									);
					/* The pe->and == NULL case is taken care of in the next recursive call when we check *proot == NULL at the top */
					}
				}
				/* not reached */
			}
		}
		else if (ret == XPF_DAGS_SAME) {

		identicalD:
			NOPRINT(xpfPrintf("\t\tXPF_DAGS_SAME adjust\n"));
			ret = FALSE;	/* no readjustments done to filterlist structure yet: also see the readjust macro */
			(*destfilter)->refcount -- /* -= xpfAbs(srcfilter->refcount) */;

		restisidenticalD:
			if ((*destfilter)->refcount <= 0) {
				XPFFilter *temp = *destfilter;

				*destfilter = (*destfilter)->next; xpfFree((char *)temp, sizeof(XPFFilter)); ret = TRUE;/* readjusted */
			}

			if (srcfilter->next == NULL) {
				if (srcpattern->next == NULL) {
					/*
					 * Delete each field one by one. Works since program are not ordered by
					 * splitting since they don't have an associated value.`
					 */

					NOPRINT(xpfPrintf("\t\tDELETE dynlist\n");)
					 d = srcpattern->programlist;
					 while (d!= NULL) {
#if	COUNTUNIT
					 	if (d->extractor.selector.countunit <= 0) goto endloop4;
#endif	COUNTUNIT
					 	pd = &(*proot)->common.pattern.programlist;
					 	while (*pd != NULL) {
							if (((*pd)->extractor.selector.level != d->extractor.selector.level) ||
								(( (*pd)->extractor.selector.offset XPF_USER_GETCOUNTUNIT((*pd))) !=
							         ( d->extractor.selector.offset XPF_USER_GETCOUNTUNIT(d)) ))
								pd = &(*pd)->next;
							else {
								if (*(unsigned long *)(*pd)->extractor.mask != *(unsigned long *)d->extractor.mask)
									pd = &(*pd)->next;
								else {
									(*pd)->refcount -- /* -= xpfAbs(d->refcount) */;
									if ((*pd)->refcount <= 0) {
										XPFProgram *temp = *pd; *pd = (*pd)->next; xpfFree((char *)temp, sizeof(XPFProgram));
									}
									goto endloop4;
								}
							}
						}

					endloop4:
						d = d->next;
					}
					READJUST_AFTER_DELETE_FILTER();
					return TRUE;
				}
				else {
					/*
					 * The idea is to find if readjustments happened only to destfilter (oldret) or to proot's pattern as well (ret)
					 * Note that if there hae been or readjustments to destfilter, there cannot be any readjustments to proot's pat.
					 */
					int oldret = ret;

					ret = FALSE;
					READJUST_AFTER_DELETE_FILTER();
					if (!oldret) return xpfDeletePatternList(srcpattern->next,
										srcpattern->next->filterlist,
										proot,
										&(*destfilter)->next,
										granularity,
										XPF_DAGS_SRCNEWDESTOLD
#if	MERGEINFO
										,deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
										,deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
										);
					else if(!ret && oldret) return xpfDeletePatternList(srcpattern->next,
											srcpattern->next->filterlist,
											proot,
											destfilter,
											granularity,
											XPF_DAGS_SRCNEWDESTOLD
#if	MERGEINFO
											,deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
											,deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
											);
					else return xpfDeletePatternList(srcpattern->next,
									srcpattern->next->filterlist,
									proot,
									destfilter,
									granularity,
									XPF_DAGS_SRCNEWDESTNEW
#if	MERGEINFO
									,deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
									,deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
									);
				}
			}
			else {
				int oldret = ret;

				ret = FALSE;
				READJUST_AFTER_DELETE_FILTER();
				if (!oldret) return xpfDeletePatternList(srcpattern,
									srcfilter->next,
									proot,
									&(*destfilter)->next,
									granularity,
									XPF_DAGS_SRCOLDDESTOLD
#if	MERGEINFO
									,deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
									,deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
									);
				else if (!ret && oldret) return xpfDeletePatternList(srcpattern,
											srcfilter->next,
											proot,
											destfilter,
											granularity,
											XPF_DAGS_SRCOLDDESTOLD
#if	MERGEINFO
											,deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
											,deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
											);
				else {
					srcpattern->header.headertype &= ~XPF_HEADER_CURRENT;	/* proot's pat is no longer current */
					return xpfDeletePatternList(srcpattern,
								srcfilter->next,
								proot,
								destfilter,
								granularity,
								XPF_DAGS_SRCNEWDESTNEW
#if	MERGEINFO
								,deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
								,deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
								);
				}
			}
			/* not reached */
		} /* end XPF_DAGS_SAME */
		else if (ret == XPF_DAGS_ERROR) return FALSE;
		else {
		notidenticalD:
			if (commonprefix > 0) {	/* handled differently even if XPF_DAGS_AFTER/BEFORE is known since keys themselves can change */
				NOPRINT(xpfPrintf("\t\tCOMPREFIX: create to hold\n");)
				location = 0;	/* both by default are OLD */

				if (destbytesleft > 0) {
					if ((*destfilter)->refcount <= 1 /* xpfAbs(srcfilter->refcount) */) {	/* obliterate the common portion */
						xpfMemcpy((*destfilter)->value, (*destfilter)->value + commonprefix, destbytesleft);
						xpfMemcpy((*destfilter)->extractor.mask, (*destfilter)->extractor.mask + commonprefix, destbytesleft);
						(*destfilter)->extractor.datatype = destbytesleft;
						(*destfilter)->extractor.selector.offset =
							(*destfilter)->extractor.selector.offset XPF_USER_GETCOUNTUNIT((*destfilter)) + commonprefix;
#if	COUNTUNIT
						(*destfilter)->extractor.selector.countunit = 1;
#endif	COUNTUNIT
						/* need not advance destfilter since in next recursive call, srcfilter&destfilter will differ w/o comprefx */
					}
					else {
						XPFFilter *f = (XPFFilter *)xpfMalloc(sizeof(XPFFilter));
						f->next = (*destfilter)->next;
						(*destfilter)->next = f;

						xpfMemcpy(f->value, (*destfilter)->value + commonprefix, destbytesleft);
						xpfMemcpy(f->extractor.mask, (*destfilter)->extractor.mask + commonprefix, destbytesleft);
						f->extractor.datatype = destbytesleft;
						f->extractor.selector.offset =
							(*destfilter)->extractor.selector.offset XPF_USER_GETCOUNTUNIT((*destfilter)) + commonprefix;
#if	COUNTUNIT
						f->extractor.selector.countunit = 1;
#endif	COUNTUNIT
						f->extractor.selector.level = 0;
						f->refcount = (*destfilter)->refcount;

						xpfBzero((char *)((*destfilter)->value + commonprefix), destbytesleft);
						xpfBzero((char *)((*destfilter)->extractor.mask + commonprefix), destbytesleft);
						(*destfilter)->extractor.datatype = commonprefix;
						(*destfilter)->refcount -- /* -= xpfAbs(srcfilter->refcount) */;
						destfilter = &((*destfilter)->next);	/* next match should begin at f */
					}
				}
				else {
					(*destfilter)->refcount -- /* -= xpfAbs(srcfilter->refcount) */;
					if ((*destfilter)->refcount <= 0) {
						XPFFilter *f = *destfilter;
						
						*destfilter = (*destfilter)->next;
						xpfFree((char *)f, sizeof(XPFFilter));
						ret = FALSE;
						READJUST_AFTER_DELETE_FILTER();
						if (ret) location |= XPF_DAGS_DESTISNEW;
					}
					else destfilter = &(*destfilter)->next;
				}

				if (srcbytesleft > 0) {
					xpfMemcpy(srcfilter->value, srcfilter->value + commonprefix, srcbytesleft);
					xpfMemcpy(srcfilter->extractor.mask, srcfilter->extractor.mask + commonprefix, srcbytesleft);
					srcfilter->extractor.datatype = srcbytesleft;
					srcfilter->extractor.selector.offset =
						srcfilter->extractor.selector.offset XPF_USER_GETCOUNTUNIT(srcfilter) + commonprefix;
#if	COUNTUNIT
					srcfilter->extractor.selector.countunit = 1;
#endif	COUNTUNIT
				}
				else srcfilter = srcfilter->next;

				if (srcfilter == NULL) {
					if (srcpattern->next == NULL) return TRUE;
					else return xpfDeletePatternList(srcpattern->next,
									srcpattern->next->filterlist,
									proot,
									destfilter,
									granularity,
									location | XPF_DAGS_SRCISNEW
#if	MERGEINFO
									,deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
									,deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
									);
				}
				else return xpfDeletePatternList(srcpattern,
								srcfilter,
								proot,
								destfilter,
								granularity,
								location
#if	MERGEINFO
								,deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
								,deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
								);
			}
			else if (ret == XPF_DAGS_AFTER) {
			case2D:
				NOPRINT(xpfPrintf("\t\tXPF_DAGS_AFTER: advance\n");)
				if ((*proot)->or == NULL) return FALSE;
				else return xpfDeletePatternList(srcpattern,
								srcfilter,
								&(*proot)->or,
								&((*proot)->or->common.pattern.filterlist),
								granularity,
								XPF_DAGS_SRCOLDDESTNEW
#if	MERGEINFO
								,deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
								,deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
								);
			}
			else {	/* neither SAME nor AFTER: => not found */
			case3D:
				return FALSE;
			}
		}
		/* not reached */
	}
	else if ( (location == XPF_DAGS_SRCNEWDESTOLD) ||
		  ((location == XPF_DAGS_SRCNEWDESTNEW) && ((*proot)->common.pattern.header.headertype & XPF_HEADER_CURRENT)) )
	{
		/*
		 * CASE 2: srcfilter is in a new header, destfilter is still in the old header.
		 * Current protocol always has a higher priority over another protocol.
		 * Control comes here also when srcpattern (as a whole) is AFTER proot's pat.
		 */
		NOPRINT(xpfPrintf("\tSRC=new, DEST=old\n");)
		if ((*proot)->or == NULL) return FALSE;
		return xpfDeletePatternList(srcpattern,
					srcfilter,
					&(*proot)->or,
					&((*proot)->or->common.pattern.filterlist),
					granularity,
					XPF_DAGS_SRCNEWDESTNEW
#if	MERGEINFO
					,deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
					,deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
					);
		/* not reached */
	}
	else if ( (location == XPF_DAGS_SRCOLDDESTNEW) ||
		  ((location == XPF_DAGS_SRCNEWDESTNEW) && (srcpattern->header.headertype & XPF_HEADER_CURRENT)) )
	{
		/*
		 * CASE 3: srcfilter is not in a new header, but destfilter is
		 * so srcpattern cannot be found in the destination in anyway.
		 */
		NOPRINT(xpfPrintf("\tSRC=old, DEST=new\n");)
		return FALSE;
	}
	else {
		/*
		 * CASE 4: Both are in new headers. The headers must match exactly. Should
		 * Control comes here also when srcpattern (as a whole) is BEFORE proot's pat.
		 */
		NOPRINT(xpfPrintf("\tNEW headers\n");)
		if (srcpattern->header.headertype != (*proot)->common.pattern.header.headertype)
			return FALSE;
#if	VARHEADER
		if (srcpattern->header.headertype & XPF_HEADER_VARIABLE) {
			if (bcmp(srcpattern->header.size.variable, (*proot)->common.pattern.header.size.variable, sizeof(XPFVariableHeader)) != 0) {
				xpfPrintf("Attempt to change header of pattern!\n");
				return FALSE;
			}
		}
		else
#endif	VARHEADER
		if (srcpattern->header.size.fixed != (*proot)->common.pattern.header.size.fixed)
			return FALSE;
		else goto case1D;
	}
}

#if	DELAYEDMSGS && STOREMSGS
#if	MERGEMSGS
#define CHECK_AND_DELETE_HASHTABLE()\
{\
	hashtable = (HashEntry **)*( unsigned long *)(*ps)->value;\
	for (i=0; i<XPF_DAGS_HASHTABLESIZE; i++) {\
		pe = &hashtable[i];\
		while (*pe != NULL) {\
			if ((currentprocessnum < 0) || ((*pe)->common.hash.processnum <= currentprocessnum - alarm)) {\
				deletedhashentry = *pe;\
				*pe = (*pe)->or;\
				(*ps)->refcount -= deletedhashentry->refcount;\
				pi = &deletedhashentry->info;\
				while (*pi != NULL) pi = &(*pi)->next;\
				*pi = *deletedinfo;\
				*deletedinfo = deletedhashentry->info;\
				deletedhashentry->info = XPF_INFO_NULL;\
				if (deletedhashentry->and != NULL)\
					xpfTerminateProcesses(&deletedhashentry->and,\
							currentprocessnum,\
							alarm,\
							granularity,\
							deletedinfo,\
							deletedmessagelist);\
				xpfFree((char *)deletedhashentry, sizeof(HashEntry));\
			}\
			else pe = &(*pe)->or;\
		}\
	}\
}
#else	MERGEMSGS
#define CHECK_AND_DELETE_HASHTABLE()\
{\
	hashtable = (HashEntry **)*( unsigned long *)(*ps)->value;\
	for (i=0; i<XPF_DAGS_HASHTABLESIZE; i++) {\
		pe = &hashtable[i];\
		while (*pe != NULL) {\
			if ((currentprocessnum < 0) || ((*pe)->common.hash.processnum <= currentprocessnum - alarm)) {\
				deletedhashentry = *pe;\
				*pe = (*pe)->or;\
				(*ps)->refcount -= deletedhashentry->refcount;\
				deletedhashentry->info = XPF_INFO_NULL;\
				if (deletedhashentry->and != NULL)\
					xpfTerminateProcesses(&deletedhashentry->and,\
							currentprocessnum,\
							alarm,\
							granularity,\
							deletedmessagelist);\
				xpfFree((char *)deletedhashentry, sizeof(HashEntry));\
			}\
			else pe = &(*pe)->or;\
		}\
	}\
}
#endif	MERGEMSGS
#else	DELAYEDMSGS && STOREMSGS
#if	MERGEINFO
#define CHECK_AND_DELETE_HASHTABLE()\
{\
	hashtable = (HashEntry **)*( unsigned long *)(*ps)->value;\
	for (i=0; i<XPF_DAGS_HASHTABLESIZE; i++) {\
		pe = &hashtable[i];\
		while (*pe != NULL) {\
			if ((currentprocessnum < 0) || ((*pe)->common.hash.processnum <= currentprocessnum - alarm)) {\
				deletedhashentry = *pe;\
				*pe = (*pe)->or;\
				(*ps)->refcount -= deletedhashentry->refcount;\
				pi = &deletedhashentry->info;\
				while (*pi != NULL) pi = &(*pi)->next;\
				*pi = *deletedinfo;\
				*deletedinfo = deletedhashentry->info;\
				deletedhashentry->info = XPF_INFO_NULL;\
				if (deletedhashentry->and != NULL)\
					xpfTerminateProcesses(&deletedhashentry->and,\
							currentprocessnum,\
							alarm,\
							granularity,\
							deletedinfo);\
				xpfFree((char *)deletedhashentry, sizeof(HashEntry));\
			}\
			else pe = &(*pe)->or;\
		}\
	}\
}
#else	MERGEINFO
#define CHECK_AND_DELETE_HASHTABLE()\
{\
	hashtable = (HashEntry **)*( unsigned long *)(*ps)->value;\
	for (i=0; i<XPF_DAGS_HASHTABLESIZE; i++) {\
		pe = &hashtable[i];\
		while (*pe != NULL) {\
			if ((currentprocessnum < 0) || ((*pe)->common.hash.processnum <= currentprocessnum - alarm)) {\
				deletedhashentry = *pe;\
				*pe = (*pe)->or;\
				(*ps)->refcount -= deletedhashentry->refcount;\
				deletedhashentry->info = XPF_INFO_NULL;\
				if (deletedhashentry->and != NULL)\
					xpfTerminateProcesses(&deletedhashentry->and,\
							currentprocessnum,\
							alarm,\
							granularity);\
				xpfFree((char *)deletedhashentry, sizeof(HashEntry));\
			}\
			else pe = &(*pe)->or;\
		}\
	}\
}
#endif	MERGEINFO
#endif	DELAYEDMSGS && STOREMSGS

/*
 * Terminates all processes with a process number < current process number - alarm.
 * Does not bother about reference counts since the portions common to > 1 process
 * get the process number of the most recent process. Since the extractors of all
 * processes generated by a pattern will be identical and values are ignored in
 * this function (processnum-s are used instead), there is no need for any
 * source-pattern or filterlist to guide us to the "proper" position: we already
 * BEGIN at the proper position (see Launch). This procedure "terminates" as soon
 * as it sees a proot's pattern that is NOT a process (implying that we have come
 * to the common-suffix of the generator pattern and its processes).
 *
 * It returns the deleted messages (not NULL only if error) in deletedmessages
 * and the deletedinfo corr. to the infos mapped to the dead processes
 * (both of which must be initialized to NULL BEFORE call).
 *
 * Assumes that if a node is not NULL, it has atleast one filter in its filterlist.
 * Maintains the invariant that the * parameter *proot can never be a HASHENTRY
 * type PATNode (simpler deletion logic). Moreover, even if the "matched node" is
 * a hash-node, if is sufficient to pass the containing (normal) pattern and the
 * corr. node since this pattern has a processnum = max(processnum of all hash nodes)
 * (see Insert and Launch).
 */
void
xpfTerminateProcesses(proot, currentprocessnum, alarm, granularity
#if	MERGEINFO
			, deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
			, deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
			)
	PATNode		**proot;
	int		currentprocessnum;
	int		alarm;
	int		granularity;
#if	MERGEINFO
	XPFInfoList	**deletedinfo;
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
	XPFMessageList	**deletedmessagelist;
#endif	DELAYEDMSGS && STOREMSGS
{
	HashEntry	**pe, *deletedhashentry = NULL, **hashtable;	/* for CHECK_AND.. */
	XPFFilter	**ps, *deletedfilter;
#if	MERGEINFO
	XPFInfoList	**pi;
#endif	MERGEINFO
	int		i, ret;	/* for CHECK_AND.. */

	DOPRINT(xpfPrintf("In TerminateProcesses\n");)
	if (*proot == NULL) return;

	/* FAULT TOLERANCE: WILL THIS EVER HAPPEN? IT CAN'T IF DELETE IS WORKING OK */
	if ((*proot)->common.pattern.filterlist == NULL) {
		DOPRINT(xpfPrintf("\tGoto NEW proot\n");)
		if ((*proot)->and == NULL) {
			xpfFree(*proot, sizeof(PATNode));
			*proot = NULL;
			return;
		}
		else {
			xpfTerminateProcesses(&((*proot)->and),
						currentprocessnum,
						alarm,
						granularity
#if	MERGEINFO
						,deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
						,deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
						);
			return;
		}
	}
	ERRPRINT(xpfAssert(((*proot)->common.pattern.filterlist != NULL) && (*proot != NULL)));
	if (!((*proot)->common.pattern.flags & XPF_FLAG_PROCESS)) {	/* return at the first absence of this flag! */
		DOPRINT(xpfPrintf("Encountered pattern!\n");)
		return;
	}

	if ((currentprocessnum < 0) || ((*proot)->common.pattern.processnum <= currentprocessnum - alarm)) {
		/* brutally delete each filter in the pattern (recursive when multivalued filter) */
		DOPRINT(xpfPrintf("\tBrutal delete of pattern: processnum=%x\n", (*proot)->common.pattern.processnum);)
		ps = &(*proot)->common.pattern.filterlist;
		while (*ps != NULL) {
			if ((*ps)->extractor.datatype < 0) {
				CHECK_AND_DELETE_HASHTABLE();
			}
			deletedfilter = *ps;
			*ps = (*ps)->next;
			xpfFree((char *)deletedfilter, sizeof(XPFFilter));
		}
	}
	else {	/* Look at each filter in the pattern: if we see a multivalued-filter, then compare the processnum of the hash-nodes and delete if necessary */
		DOPRINT(xpfPrintf("\tLook and delete pattern: processnum=%x\n", (*proot)->common.pattern.processnum);)
		ps = &(*proot)->common.pattern.filterlist;
		while (*ps != NULL) {
			if ((*ps)->extractor.datatype >= 0) {
				ps = &(*ps)->next;
				continue;
			}
			else CHECK_AND_DELETE_HASHTABLE();
			if ((*ps)->refcount <= 0) {
				deletedfilter = *ps;
				*ps = (*ps)->next;
				xpfFree((char *)deletedfilter, sizeof(XPFFilter));
			}
			else ps = &(*ps)->next;
		}
	}

	if ((*proot)->common.pattern.filterlist == NULL) {
		READJUST_AFTER_DELETE_PATTERN();
		if (*proot == NULL) return;
	}
	/* else everything is in the proper place */

	xpfTerminateProcesses(proot,
				currentprocessnum,
				alarm,
				granularity
#if	MERGEINFO
				,deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
				,deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
				);
	return;
}

/*
 * Returns XPF_USER_NEWPATTERN if at least one processe was launched due to
 * some matched node along the path taken by the message, XPF_USER_OLDPATTERN
 * if no new processes were launched (state unchanged), and XPF_USER_FAILURE
 * if there was an error in following the path of the message.
 *
 * A Process is generated from a programlist in a pattern in a matched node and
 * is inserted into the DAG such that its prefix is the same as the prefix of
 * the pattern generating it, but its suffix is NULL since processes act as a
 * cache of the longest matching path. So, if the lowest protocol can be made to
 * launch a process, then it is best since its prefix will be the smallest (=0).
 *
 * This function also sets "pinfo" to the information mapped to the longest
 * matching node (if that node launches a process, this information is the one
 * mapped to the process, NOT to the node!), and "pdeletedmessagelist" to the
 * delayed messages at the matched nodes (after removing them from these nodes).
 */
int
xpfLaunchProcesses(state, pinfo
#if	MERGEINFO
		, pdeletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
		, pdeletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
		)
	XPFState	*state;
#if	MERGEINFO
	XPFInfoList	**pinfo;
#else	MERGEINFO
	XPFInformation	*pinfo;
#endif	MERGEINFO
#if	MERGEINFO
	XPFInfoList	**pdeletedinfo;
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
	XPFMessageList	**pdeletedmessagelist;
#endif	DELAYEDMSGS && STOREMSGS
{
	PATNode		**proot;			/* root at which the newly created process should be inserted, NOT NECESSARILY MatchedNodes[i]! */
	PATNode		**curnode;			/* points to the node from which the programs should be taken, NOT NECESSARILY proot! */
	XPFPattern	*curpat;			/* points to the pattern that is launching the process, NOT NECESSARILY (*proot)->common.pattern! */
	XPFPattern	*new;
	XPFFilter	**ps;
	XPFProgram	*d;
	register int	i, k;
	int		numpat;
	int		ret = 0;
	PATNode		*newnode;
#if	MERGEINFO
	XPFInfoList	*deletedinfo, **pl, *l = XPF_INFO_NULL;
#else	MERGEINFO
	XPFInformation	l = XPF_INFO_NULL;
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
	XPFMessageList	*deletedmessagelist, **pm, **m;
#endif	DELAYEDMSGS && STOREMSGS
	int		willfork = FALSE;		/* will current node launch a program even if it doesn't have one (prev. node = hash-node had one) */
	int		extract_next_time = FALSE;	/* this is valid ACROSS nodes since a new node might start with a current/new pattern */
	int		copy, copytype;			/* tells when to use current header and when to start new one in process (set if proot is changed) */
	int		must_continue;			/* tells if we need to just un-extract headers or launch too */

	DOPRINT(xpfPrintf("LaunchPrograms: numnodes = %x\n", state->xpfMatchedNum);)

#if	DELAYEDMSGS && STOREMSGS
	*pdeletedmessagelist = NULL;
#endif	DELAYEDMSGS && STOREMSGS
	if ((state->xpfMatchedNum <= 0) || (state->xpfMatchedNum > XPF_DAGS_MAXDEPTH)) return XPF_USER_FAILURE;
	ERRPRINT(xpfAssert(state->xpfHeaderCurrent == state->xpfHeaderNum));

	/*
	 * Go in the loop from the last matched node to the first matched node, i.e.,
	 * from the most specific node to the least specific node, and launch
	 * programs on the way, i.e., modify the DAG from the leaf up to the root.
	 */
	*pinfo = (*state->xpfMatchedNodes[state->xpfMatchedNum-1])->info;	/* If programs are generated, then assign the new info later below */
	DOPRINT(xpfPrintf("MatchedNum = %x lastinfo = %x\n", state->xpfMatchedNum, *pinfo);)

	for(i=state->xpfMatchedNum-1; i>=0; i--) {
		must_continue = FALSE;
		if ((i > 0) && ((*state->xpfMatchedNodes[i])->nodetype & XPF_DAGS_HASHENTRY)) {
			if ((*state->xpfMatchedNodes[i])->common.hash.programlist != NULL) willfork = TRUE;	/* used in next iteration below */
			continue;	/* can't set must continue since code below assumes current matched node is never a hash entry */
		}
		/*
		 * Process nodes are accessed and handled internally; plus I will never miss something from the
		 * hash-node since the parent normal node will never have any programs: taken over by 1st #entry.
		 */
		if ( ((*state->xpfMatchedNodes[i])->nodetype & XPF_DAGS_PROCESS) ||	/* do nothing: just unextract */
		     (((*state->xpfMatchedNodes[i])->common.pattern.programlist == NULL) && !willfork) ) must_continue = TRUE;
		willfork = FALSE;

		/*
		 * The above takes care of all special cases. From this point on, we are sure
		 * that unless there is an insertion/overflow error, a processes will be launched.
		 */

		/*
		 * The root at which insertions and deletions should begin: NEVER a HASHENTRY + CAN CHANGE ONLY IF the "last" process gets killed.
		 * This cannot be state->xpfMatchedNodes[i] itself since that pattern might be at the end of a sequential OR of patterns and we
		 * want the new process to be inserted BEFORE any such sequence of patterns (since processes must be matched BEFORE corr. patterns).
		 */
		proot = ((i>0) ? &(*state->xpfMatchedNodes[i-1])->and : &state->xpfAssociationSet);
		curpat = &(*state->xpfMatchedNodes[i])->common.pattern;

		copy = 0;
		if (extract_next_time) {
			if (--state->xpfHeaderCurrent <= 0) {
				state->xpfErrno = - XPF_DAGS_MAXDEPTH;
				return XPF_USER_FAILURE;	/* Can't happen if Get was ok: error recorvery */
			}
			extract_next_time = FALSE;
		}

		if (!(curpat->header.headertype & XPF_HEADER_CURRENT)) {
			extract_next_time = TRUE;
			if (must_continue) continue;	/* don't do anything other than header un-extraction */
		}
		else {	/* CURRENT_HEADER */
			if (must_continue) continue;	/* don't do anything other than header un-extraction */
			if (i>0) {
				/*
				 * It is guaranteed that these two calculations will never yield a hashentry (since process/current flags are
				 * same in both hashentries and their parent nodes, so that we never quit from the while loop in the middle).
				 */
				/* must to go new proot if prev node was a process since I must insert at the first node among the processes */
				if ((*state->xpfMatchedNodes[i-1])->nodetype & XPF_DAGS_PROCESS) {
					k = i-2;
					while ((k>0) && ((*state->xpfMatchedNodes[k])->nodetype & XPF_DAGS_PROCESS)) k--;
					/* insert into the first matched process node */
					if (k > 0) proot = &(*state->xpfMatchedNodes[k])->and;
					else proot = &state->xpfAssociationSet;
					/* can't say i = k; since we must UNEXTRACT the headers corr. to the patterns just traversed */
					copy = 1;
					copytype = (*proot)->common.pattern.header.headertype;
				}
				/* continue further back until I come to a node that is not current and insert programs right after it (into its and) */
				else {
					k = i-1;
					while ((k>0) && ((*state->xpfMatchedNodes[k])->common.pattern.header.headertype & XPF_HEADER_CURRENT)) k--;
					/* insert into the first matched non-current header node */
					if (k > 0) proot = &(*state->xpfMatchedNodes[k])->and;
					else proot = &state->xpfAssociationSet;
					/* can't say i = k; since the skipped node's patterns might launch programs in future iterations */
					copy = 1;
					copytype = (*proot)->common.pattern.header.headertype;
				}
			}
			else proot = &state->xpfAssociationSet;	/* nowhere else to go */
		}

		/*
		 * Extract program (newpattern) from the "correct" programlist. Also, remember that
		 * if no programs here and this is the last pattern and the next node was a hash node,
		 * then use the program of the hash node. This is independent of proot calculation.
		 */
		l = *pinfo;
		curnode = state->xpfMatchedNodes[i];
		d = curpat->programlist;
#if	DELAYEDMSGS && STOREMSGS
		m = &(*curnode)->messagelist;
#endif	DELAYEDMSGS && STOREMSGS
		if ((d == NULL) &&
		    (i+1 < state->xpfMatchedNum) &&
		    ((*state->xpfMatchedNodes[i+1])->nodetype & XPF_DAGS_HASHENTRY) ) {
			curnode = &(*state->xpfMatchedNodes[i+1]);
			d = (*curnode)->common.hash.programlist;
#if	DELAYEDMSGS && STOREMSGS
			m = &(*curnode)->messagelist;
#endif	DELAYEDMSGS && STOREMSGS
		}

		NOWPRINT(xpfPrintProgramList(d); printf("\n");)
		if (d == NULL) {
			newnode = *curnode;	/* don't have to launch anything, just set the node for information retrieval */
		}
		else if (curpat->alarm <= 0) {	/* launching of processes disabled since 0 concurrent processes; delete all old processes */
			newnode = *curnode;
			if (++state->xpfClock >= state->xpfTimer) {
				state->xpfClock = 0;
				NOWPRINT(xpfPrintf("LaunchProcesses: terminating all old processes: curprocess = %x\n", curpat->processnum);)
				xpfTerminateProcesses(proot,
							curpat->processnum,
							0,
							state->xpfGranularity
#if	MERGEINFO
							,pdeletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
							,pdeletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
							);
				DOPRINT(xpfPrintPATTree(*proot, 0);)
			}
			ret |= XPF_USER_TERMINATED;
		}
		else {	/* launch new process after deleting all those that have lived too long if Timer expires */
			DOPRINT(xpfAssert(*proot != NULL));
			curpat->processnum ++;
			if ((*curnode)->nodetype & XPF_DAGS_HASHENTRY) (*curnode)->common.hash.processnum = curpat->processnum;

			if (++state->xpfClock >= state->xpfTimer) {
				state->xpfClock = 0;
				NOWPRINT(xpfPrintf("LaunchProcesses: terminating aged (>%x) processes: curprocess = %x\n", curpat->alarm, curpat->processnum);)
				xpfTerminateProcesses(proot,
						curpat->processnum,
						curpat->alarm,
						state->xpfGranularity
#if	MERGEINFO
						,pdeletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
						,pdeletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
						);
				DOPRINT(xpfPrintPATTree(*proot, 0);)
			}

			DOPRINT(xpfPrintf("LaunchProcess: creating process\n");)
			new = (XPFPattern *)xpfMalloc(sizeof(XPFPattern));
			(new)->flags = 0;
			(new)->processnum = curpat->processnum;
			(new)->alarm = 0;	/* not used in processes */
#if	VARHEADER
			(new)->header.size.variable = NULL;
#endif	VARHEADER
			COPY_HEADER(&((new)->header), &(curpat->header));
			if (copy) new->header.headertype = copytype;
			ps = &(new)->filterlist;
			while (d != NULL) {
				*ps = (XPFFilter *)xpfMalloc(sizeof(XPFFilter));
				xpfMemcpy(&((*ps)->extractor), &(d->extractor), sizeof(XPFExtractor));
				memmaskcpy((*ps)->value,
					   state->xpfHeaderData[state->xpfHeaderCurrent - 1] +
						d->extractor.selector.offset XPF_USER_GETCOUNTUNIT(d),
					   d->extractor.mask,
					   XPF_DATA_MAXSIZE);
				(*ps)->refcount = 1 /* xpfAbs(d->refcount) */;
				ps = &(*ps)->next;
				d = d->next;
			}
			*ps = NULL;
			(new)->programlist = NULL;	/* programlist of processes is always NULL */
			(new)->next = NULL;
			(new)->flags |= XPF_FLAG_PROCESS;
			DOPRINT(xpfAssert((new)->filterlist != NULL));

			/*
			 * This is very similar to xpfSetupAssociationSuffix but in this case the prefix
			 * is already known to be proot, and the suffixnode is known to be NULL (cache).
			 * Moreover, InsertPatternList MUST be called with XPF_FLAG_DAGS (not USER!).
			 */
			DOPRINT(xpfPrintf("LaunchProcesses: nodenum=%x\n", i);)
			DOPRINT(xpfAssert((new)->filterlist != NULL));

			DOPRINT(xpfPrintf("node generating program: --\n");)
			DOPRINT(xpfPrintPATTree((*curnode), 0);)

			if ((newnode = xpfInsertPatternList(new,
						new->filterlist,
						proot,
						&(*proot)->common.pattern.filterlist,
						state->xpfGranularity,
						XPF_DAGS_SRCNEWDESTNEW,
						XPF_FLAG_DAGS)) == NULL) {
				ERRPRINT(xpfPrintf("Could not launch process!\n"));
			}
			else {
				/*
				 * Copy the information in l into newnode (process has a copy of info at longest matching pattern).
				 * Since we go backward, it's okay to forget about the node with the "source" info once we
				 * copy it. Moreover, "suffixnode" can ALWAYS be taken to be NULL. Since we use
				 * a DAG, even if (*curnode) changes after InsertPatternList, curnode ITSELF CANNOT CHANGE.
				 */
#if	MERGEINFO
				pl = &(newnode)->info;
				while (l != NULL) {
					*pl = (XPFInfoList *)xpfMalloc(sizeof(XPFInfoList));
					(*pl)->information = l->information;
					l = l->next;
					pl = &(*pl)->next;
				}
				*pl = NULL;
#else	MERGEINFO
				newnode->info = l;
#endif	MERGEINFO
				newnode->and = NULL;

#if	DELAYEDMSGS && STOREMSGS
				/*
				 * Merge the delayed messagelist.
				 */
				pm = m;
				while(*m != NULL) m = &(*m)->next;
				*m = *pdeletedmessagelist;
				*pdeletedmessagelist = *pm;
				*pm = NULL;
#endif	DELAYEDMSGS && STOREMSGS

				DOPRINT(xpfPrintf("inserted program into: --\n");)
				DOPRINT(xpfPrintPATTree(newnode, 0);)

#if	MERGEINFO
				/* Ignore the deletedinfo and deletedmessagelist: they will be NULL after return */
				deletedinfo = XPF_INFO_NULL;
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
				deletedmessagelist = NULL;
#endif	DELAYEDMSGS && STOREMSGS

				DOPRINT(xpfPrintf("after suffix setup: --\n");)
				DOPRINT(xpfPrintPATTree(newnode, 0);)
				ret |= XPF_USER_LAUNCHED;
				if ((i == state->xpfMatchedNum - 1) ||
				    ((i == state->xpfMatchedNum - 2) && ((*state->xpfMatchedNodes[i+1])->nodetype & XPF_DAGS_HASHENTRY))
				   ) {
#if	MERGEINFO
					/* must assign the information to be returned on lookup: pointer to ANOTHER list. */
					*pinfo = newnode->info;
					printf("launch: process's info=%x\n", *pinfo);
					DOPRINT(xpfPrintf("Got match %x after trying launch at i=%x\n", newnode->info, i);)
#endif	MERGEINFO
					ret |= XPF_USER_MATCHEDNEW;
				}
			}

			/* cleanup after trying to launch */
			xpfDestroyPatternList(new);
		}
	}

	return ret;
}
