/*****************************************************************************
 * xpf_driver.c: Contains functions that can be directly plugged-into a	     *
 *		a Unix-style driver. Delayed messages (OOO delivery) aren't  *
 *		handled. Uses routines in xpf_interface.c. For an x-Kernel   *
 *		style driver, one just has to write xpf_demux using xpf_read.*
 *		Delivery and Xmission of messages are done outside xpf.	     *
 *****************************************************************************/

#include <stdio.h>
#include "xpf_driver.h"

#define DOPRINT(s)
#define NOPRINT(s)

/*
 * On allocation error, it returns NULL; otherwise, it returns an
 * object-descriptor which can be used in the functions below.
 */

XPFObject
xpf_open()
{
	XPFObject	xpfobj;

	if (xpfCreateObject(&xpfobj)) return xpfobj;
	return NULL;
}

/*
 * Returns TRUE on success, FALSE on failure.
 */

int
xpf_ioctl(xpfobj, opcode, buf, len)
	XPFObject	xpfobj;
	int		opcode;
	char		*buf;
	int		len;
{
	int		retval;
#if	MERGEINFO
	XPFInfoList	info;
	XPFInfoList	*deletedinfo = XPF_INFO_NULL, *temp1;
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
	XPFMessageList	*deletedmessagelist = NULL;
	XPFMessageList	*temp2;
#endif	DELAYEDMSGS && STOREMSGS

NOPRINT(	xpfPrintf("In xpfControl %x\n", opcode);)

	switch(opcode)
	{
	case XPF_SETUP:
	{
		if (len < sizeof(XPFSetupArg *)) return FALSE;
#if	MERGEINFO
		info.next = XPF_INFO_NULL;
		info.information = ((XPFSetupArg *)buf)->information;
#endif	MERGEINFO
		if (!xpfValidPatternList(((XPFSetupArg *)buf)->pattern)) return FALSE;
		return xpfSetupAssociation(xpfobj,
			((XPFSetupArg *)buf)->pattern
#if	MERGEINFO
			,&info
#else	MERGEINFO
			,((XPFSetupArg *)buf)->information
#endif	MERGEINFO
			);
	}
	case XPF_SETUPSUFFIX:
	{
		if (len < sizeof(XPFSetupSuffixArg *)) return FALSE;
#if	MERGEINFO
		info.next = XPF_INFO_NULL;
		info.information = ((XPFSetupSuffixArg *)buf)->information;
#endif	MERGEINFO
		if (!xpfValidPatternList(((XPFSetupSuffixArg *)buf)->prefix)) return FALSE;
		if (!xpfValidPatternList(((XPFSetupSuffixArg *)buf)->new)) return FALSE;
		if (!xpfValidPatternList(((XPFSetupSuffixArg *)buf)->findsuffix)) return FALSE;
		return xpfSetupAssociationSuffix(xpfobj,
			((XPFSetupSuffixArg *)buf)->prefix,
			((XPFSetupSuffixArg *)buf)->new,
			((XPFSetupSuffixArg *)buf)->findsuffix
#if	MERGEINFO
			,&info
#else	MERGEINFO
			,((XPFSetupSuffixArg *)buf)->information
#endif	MERGEINFO
			);
	}
	case XPF_REMOVE:
	{
#if	MERGEINFO
		deletedinfo = XPF_INFO_NULL;
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
		deletedmessagelist = NULL;
#endif	DELAYEDMSGS && STOREMSGS
		retval = xpfRemoveAssociation(xpfobj,
				((XPFRemoveArg *)buf)->pattern
#if	MERGEINFO
				,&deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
				,&deletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
				);

#if	MERGEINFO
	 	for (; deletedinfo!=XPF_INFO_NULL;
		    /* xpfDestroyInformation(deletedinfo->information), -- info is a pointer, nothing to destroy */
		    temp1 = deletedinfo,
		    deletedinfo=deletedinfo->next,
		    xpfFree((char *)temp1, sizeof(XPFInfoList)));
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
		for (; deletedmessagelist!=NULL;
		    msgDestroy(((XPFPMessage *)deletedmessagelist->message)->contents),
		    temp2 = deletedmessagelist,
		    deletedmessagelist=deletedmessagelist->next,
		    xpfFree((char *)temp2, sizeof(XPFMessageList)));
#endif	DELAYEDMSGS && STOREMSGS

		return retval;
	}
	default:
		return FALSE;
	}
}

/*
 * Returns NULL if there is no information which matches the message,
 * the mapped information otherwise.
 */

XPFInformation
xpf_read(xpfobj, msg, len)
	XPFObject	xpfobj;
	XPFMessage	msg;
	int		len;
{
#if	MERGEINFO
	XPFInfoList	*associatedinfo = XPF_INFO_NULL;
	XPFInfoList	*deletedinfo = XPF_INFO_NULL, *temp1;
#else	MERGEINFO
	XPFInformation	associatedinfo = XPF_INFO_NULL;
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
	XPFMessageList	*delayedmsgs = NULL;
#endif	DELAYEDMSGS && STOREMSGS
	int		ret;

	NOPRINT(xpfPrintf("xpfLookup: length=%x\n", len);)
	NOPRINT(
	{
		int i;
		for (i=0; i<len; i+=2)
			xpfPrintf("%x ", *((unsigned short *)(&msg[i])));
		xpfPrintf("\n");
	}
	)
	ret = xpfLookupAssociation(xpfobj, msg, len, &associatedinfo
#if	MERGEINFO
				, &deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
				, &delayedmsgs
#endif	DELAYEDMSGS && STOREMSGS
				);
	DOPRINT(xpfPrintf("lookup: length %x, returned %x, info %x\n", len, ret, associatedinfo);)

	if (
#if	DELAYEDMSGS	/* Not handling delayed messages in the driver as of now */
	    (ret & XPF_USER_DELAYED) ||
#endif	DELAYEDMSGS
	    (ret & XPF_USER_FAILURE)
	   ) {
		return XPF_INFO_NULL;
	}

#if	MERGEINFO
 	for (; deletedinfo!=XPF_INFO_NULL;
	    /* xpfDestroyInformation(deletedinfo->information), -- info is pointer, nothing to destroy */
	    temp1 = deletedinfo,
	    deletedinfo=deletedinfo->next,
	    xpfFree((char *)temp1, sizeof(XPFInfoList)));
#endif	MERGEINFO
#if	DELAYEDMSGS
	while(delayedmsgs!=NULL)
	    xpfDemux(state, ????????? /* Not handling delayed messages in the driver at present */
		      ((XPFPMessage *)delayedmsgs->message)->source,
		      ((XPFPMessage *)delayedmsgs->message)->contents);
#endif	DELAYEDMSGS

#if	MERGEINFO
	if (associatedinfo == XPF_INFO_NULL) return XPF_INFO_NULL;
	return associatedinfo->information;
#else	MERGEINFO
	return associatedinfo;
#endif	MERGEINFO
}

/*
 * Destroys the object, returns nothing.
 */

void
xpf_close(xpfobj)
	XPFObject	xpfobj;
{
#if	MERGEINFO
	XPFInfoList	*deletedinfo = XPF_INFO_NULL, *temp1;
#endif	MERGEINFO
#if	DELAYEDMSGS
	XPFMessageList	*deletedmessagelist = NULL;
	XPFMessageList	*temp2;
#endif	DELAYEDMSGS

	xpfDestroyObject(xpfobj
#if	MERGEINFO
			, &deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS
			, &deletedmessagelist
#endif	DELAYEDMSGS
			);

#if	MERGEINFO
	for (; deletedinfo!=XPF_INFO_NULL;
	    /* xpfDestroyInformation(deletedinfo->information), -- info is a pointer, nothing to destroy */
	    temp1 = deletedinfo,
	    deletedinfo=deletedinfo->next,
	    xpfFree((char *)temp1, sizeof(XPFInfoList)));
#endif	MERGEINFO
#if	DELAYEDMSGS
	for (; deletedmessagelist!=NULL;
	    /* msgDestroy(((XPFPMessage *)deletedmessagelist->message)->contents), -- message is a pointer, nothing to destroy */
	    temp2 = deletedmessagelist,
	    deletedmessagelist=deletedmessagelist->next,
	    xpfFree((char *)temp2, sizeof(XPFMessageList)));
#endif	DELAYEDMSGS

	return;
}
